# Strip Pooling: Rethinking spatial pooling for scene parsing (CVPR 2020)
import torch
from torch import nn
import torch.nn.functional as F

class StripPooling(nn.Module):
    """
    Reference:
    """
    def __init__(self, in_channels, pool_size, norm_layer=None):
        super(StripPooling, self).__init__()
        self.pool1 = nn.AdaptiveAvgPool2d(pool_size[0])
        self.pool2 = nn.AdaptiveAvgPool2d(pool_size[1])
        self.pool3 = nn.AdaptiveAvgPool2d((1, None))
        self.pool4 = nn.AdaptiveAvgPool2d((None, 1))
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d

        inter_channels = int(in_channels / 4)
        self.conv1_1 = nn.Sequential(nn.Conv2d(in_channels, inter_channels, 1, bias=False),
                                     norm_layer(inter_channels),
                                     nn.ReLU(True))
        self.conv1_2 = nn.Sequential(nn.Conv2d(in_channels, inter_channels, 1, bias=False),
                                     norm_layer(inter_channels),
                                     nn.ReLU(True))
        self.conv2_0 = nn.Sequential(nn.Conv2d(inter_channels, inter_channels, 3, 1, 1, bias=False),
                                     norm_layer(inter_channels))
        self.conv2_1 = nn.Sequential(nn.Conv2d(inter_channels, inter_channels, 3, 1, 1, bias=False),
                                     norm_layer(inter_channels))
        self.conv2_2 = nn.Sequential(nn.Conv2d(inter_channels, inter_channels, 3, 1, 1, bias=False),
                                     norm_layer(inter_channels))
        self.conv2_3 = nn.Sequential(nn.Conv2d(inter_channels, inter_channels, (1, 3), 1, (0, 1), bias=False),
                                     norm_layer(inter_channels))
        self.conv2_4 = nn.Sequential(nn.Conv2d(inter_channels, inter_channels, (3, 1), 1, (1, 0), bias=False),
                                     norm_layer(inter_channels))
        self.conv2_5 = nn.Sequential(nn.Conv2d(inter_channels, inter_channels, 3, 1, 1, bias=False),
                                     norm_layer(inter_channels),
                                     nn.ReLU(True))
        self.conv2_6 = nn.Sequential(nn.Conv2d(inter_channels, inter_channels, 3, 1, 1, bias=False),
                                     norm_layer(inter_channels),
                                     nn.ReLU(True))
        self.conv3 = nn.Sequential(nn.Conv2d(inter_channels * 2, in_channels, 1, bias=False),
                                   norm_layer(in_channels))

    def forward(self, x):
        _, _, h, w = x.size()
        x1 = self.conv1_1(x)
        x2 = self.conv1_2(x)
        x2_1 = self.conv2_0(x1)
        x2_2 = F.interpolate(self.conv2_1(self.pool1(x1)), (h, w), mode='bilinear', align_corners=True)
        x2_3 = F.interpolate(self.conv2_2(self.pool2(x1)), (h, w), mode='bilinear', align_corners=True)
        x2_4 = F.interpolate(self.conv2_3(self.pool3(x2)), (h, w), mode='bilinear', align_corners=True)
        x2_5 = F.interpolate(self.conv2_4(self.pool4(x2)), (h, w), mode='bilinear', align_corners=True)
        x1 = self.conv2_5(F.relu_(x2_1 + x2_2 + x2_3))
        x2 = self.conv2_6(F.relu_(x2_5 + x2_4))
        out = self.conv3(torch.cat([x1, x2], dim=1))
        return F.relu_(x + out)

if __name__ == '__main__':
    block = StripPooling(64, (20, 12))

    input = torch.rand(4, 64, 64, 64)
    output = block(input)
    print(input.size(), output.size())




import pyzjr.augmentation.transforms.opencv as cv2_function
import pyzjr.augmentation.transforms.pillow as pil_function
import pyzjr.augmentation.transforms.tvision as tvision_function

from .opencv import *
from .pillow import *
from ._utils import *
from .normal import *
from .tvision import *


"""
Copyright (c) 2024, Auorui.
All rights reserved.
time 2024-01-25
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from pyzjr.nn.torchutils.loss_utils import boundary_loss

__all__ = ["L1Loss", "L2Loss", "BCELoss", "CrossEntropyLoss", "FocalLoss", "DiceLoss"]

class L1Loss(nn.Module):
    """
    L1损失，也称为平均绝对误差（MAE），测量预测输出中的每个元素与目标或地面实况中的相应元素之间的平均绝对差。
    在数学上，它表示为预测值和目标值之间差异的绝对值的平均值。与L2损耗相比，L1损耗对异常值不那么敏感。依据公式实现。
    Args:
        input (torch.Tensor): The predicted output.
        target (torch.Tensor): The target or ground truth.
        reduction (str, optional): Specifies the reduction to apply to the output.
            Options are 'none', 'mean', or 'sum'. Default is 'mean'.
    Examples::
        >>> criterion1 = nn.L1Loss()
        >>> criterion2 = L1Loss()
        >>> input_data=torch.Tensor([2, 3, 4, 5])
        >>> target_data=torch.Tensor([4, 5, 6, 7])
        >>> loss1 = criterion1(input_data, target_data)  # tensor(2.)
        >>> loss2 = criterion2(input_data, target_data)  # tensor(2.)
    Returns:
        torch.Tensor: The L1 loss between input and target.
    """
    def __init__(self):
        super(L1Loss, self).__init__()

    def forward(self, input, target):
        loss = torch.mean(torch.abs(input - target))
        return loss

class L2Loss(nn.Module):
    """
    L2损失，也称为均方误差（MSE），测量预测输出中的每个元素与目标或地面实况中的相应元素之间的平均平方差。
    在数学上，它表示为预测值和目标值之间差异的平方的平均值。相比于L1损耗，L2损耗对异常值更敏感。依据公式实现。
    在torch当中是MSELoss
    Args:
        input (torch.Tensor): The predicted output.
        target (torch.Tensor): The target or ground truth.
        reduction (str, optional): Specifies the reduction to apply to the output.
            Options are 'none', 'mean', or 'sum'. Default is 'mean'.
    Examples::
        >>> criterion1 = nn.MSELoss()
        >>> criterion2 = L2Loss()
        >>> input_data=torch.Tensor([2, 3, 4, 5])
        >>> target_data=torch.Tensor([4, 5, 6, 7])
        >>> loss1 = criterion1(input_data, target_data)  # tensor(4.)
        >>> loss2 = criterion2(input_data, target_data)  # tensor(4.)

    Returns:
        torch.Tensor: The L2 loss between input and target.
    """
    def __init__(self):
        super(L2Loss, self).__init__()

    def forward(self, input, target):
        loss = torch.mean(torch.pow(input - target, 2))
        return loss

class BCELoss(nn.Module):
    """
    二元交叉熵损失（Binary Cross Entropy Loss），也称为对数损失。
    用于测量预测输出中的每个元素与目标或地面实况中的相应元素之间的对数概率差异。依据公式实现。
    Args:
        input (torch.Tensor): The predicted output.Map to (0,1) through sigmoid function.
        target (torch.Tensor): The target or ground truth.
        reduction (str, optional): Specifies the reduction to apply to the output.
            Options are 'none', 'mean', or 'sum'. Default is 'mean'.

    Examples::
        >>> criterion1 = nn.BCELoss()
        >>> criterion2 = BCELoss()
        >>> input_data = torch.randn((5,))
        >>> target_data = torch.randint(0, 2, (5,), dtype=torch.float32)
        >>> loss1 = criterion1(torch.sigmoid(input_data), target_data)
        >>> loss2 = criterion2(input_data, target_data)
        >>> print("PyTorch BCELoss:", loss1.item())
        >>> print("Custom BCELoss:", loss2.item())

    Returns:
        torch.Tensor: The binary cross entropy loss between input and target.
    """
    def __init__(self, ignore_index=None, reduction='mean'):
        super(BCELoss, self).__init__()
        self.reduction = reduction
        self.ignore_index = ignore_index

    def forward(self, input, target):
        if self.ignore_index:
            input = input[:,:self.ignore_index,...]
            target = target[:, :self.ignore_index, ...]
        input = torch.sigmoid(input)
        loss = - (target * torch.log(input) + (1 - target) * torch.log(1 - input))
        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        elif self.reduction == 'none':
            return loss

class CrossEntropyLoss(nn.Module):
    """
    交叉熵损失（Cross Entropy Loss）用于多分类问题。
    用于测量预测输出和目标分布之间的交叉熵。依据公式实现。
    Args:
        input (torch.Tensor): The predicted output (logits).
        target (torch.Tensor): The target or ground truth (class labels).
        reduction (str, optional): Specifies the reduction to apply to the output.
            Options are 'none', 'mean', or 'sum'. Default is 'mean'.

    Examples::
        >>> criterion1 = nn.CrossEntropyLoss()
        >>> criterion2 = CrossEntropyLoss()
        >>> input_data = torch.randn((3, 5), requires_grad=True)
        >>> target_data = torch.randint(0, 5, (3,))
        >>> loss1 = criterion1(input_data, target_data)
        >>> loss2 = criterion2(input_data, target_data)
        >>> print("PyTorch CrossEntropyLoss:", loss1.item())
        >>> print("Custom CrossEntropyLoss:", loss2.item())

    Returns:
        torch.Tensor: The cross entropy loss between input and target.
    """
    def __init__(self, ignore_index=None, reduction='mean'):
        super(CrossEntropyLoss, self).__init__()
        self.reduction = reduction
        self.ignore_index = ignore_index

    def forward(self, input, target):
        if self.ignore_index:
            input = input[:,:self.ignore_index,...]
            target = target[:, :self.ignore_index, ...]
        return nn.NLLLoss(reduction=self.reduction)(F.log_softmax(input, dim=1), target)


class FocalLoss(nn.Module):
    """
    Focal Loss 用于解决类别不平衡问题，通过缩小易分类的类别的损失来关注难分类的类别。依据公式实现。

    Args:
        alpha (float, optional): 控制易分类的类别的权重，大于1表示增加权重，小于1表示减小权重。默认为1.
        gamma (float, optional): 控制难分类的类别的损失的下降速度，大于0表示下降较慢，小于0表示下降较快。默认为2.
        reduction (str, optional): Specifies the reduction to apply to the output.
            Options are 'none', 'mean', or 'sum'. Default is 'mean'.

    Examples::
        >>> criterion = FocalLoss(alpha=1, gamma=2, reduction='mean')
        >>> input_data = torch.randn((5, 3), requires_grad=True)
        >>> target_data = torch.randint(0, 3, (5,))
        >>> loss = criterion(input_data, target_data)
        >>> print("Focal Loss:", loss.item())
    """
    def __init__(self, alpha=1, gamma=2, ignore_index=None, reduction='mean'):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction
        self.ignore_index = ignore_index

    def forward(self, input, target):
        if self.ignore_index:
            input = input[:,:self.ignore_index,...]
            target = target[:, :self.ignore_index, ...]
        ce_loss = F.cross_entropy(input, target, reduction='none')
        class_weights = torch.exp(-ce_loss)
        focal_loss = self.alpha * (1 - class_weights) ** self.gamma * ce_loss
        if self.reduction == 'mean':
            return focal_loss.mean()
        elif self.reduction == 'sum':
            return focal_loss.sum()
        elif self.reduction == 'none':
            return focal_loss

class DiceLoss(nn.Module):
    """
    Dice Loss 测量预测的和目标的二进制分割掩码之间的不相似性。它被计算为1减去Dice系数，Dice系数是重叠的度量
    在预测区域和目标区域之间。

    Examples::
        >>> criterion = DiceLoss(reduction='mean')
        >>> predictions = torch.rand((1,2,16,16), dtype=torch.float32)
        >>> targets = torch.randn((1,2,16,16), dtype=torch.float32)
        >>> loss = criterion(predictions, targets)
        >>> print("Dice Loss:", loss.item())

    Returns:
        torch.Tensor: The Dice Loss between input and target.
    """
    def __init__(self, ignore_index=None, reduction='mean', eps=1e-5):
        super(DiceLoss, self).__init__()
        self.eps = eps
        self.ignore_index = ignore_index
        self.reduction = reduction

    def forward(self, input, target):
        if self.ignore_index:
            input = input[:,:self.ignore_index,...]
            target = target[:, :self.ignore_index, ...]

        input = torch.sigmoid(input)
        intersection = torch.sum(input * target)
        union = torch.sum(input) + torch.sum(target)
        dice_loss = 1 - (2 * intersection + self.eps) / (union + self.eps)
        if self.reduction == 'mean':
            return dice_loss.mean()
        elif self.reduction == 'sum':
            return dice_loss.sum()
        elif self.reduction == 'none':
            return dice_loss


class BoundaryLoss(nn.Module):
    """
    计算二进制分割的边界损失

    Args:
        None

    Examples:
        >>> criterion = BoundaryLoss()
        >>> outputs_soft = torch.rand((1, 1, 3, 3))  # model prediction
        >>> outputs_soft = torch.sigmoid(outputs_soft)
        >>> label_batch = torch.randint(2, (1, 1, 3, 3))  # binary segmentation mask
        >>> loss = criterion(outputs_soft, label_batch)
        >>> print("Boundary Loss:", loss.item())
    Returns:
        torch.Tensor: The Boundary Loss between model predictions and ground truth.
    """
    def __init__(self, ignore_index=None, reduction='mean'):
        super(BoundaryLoss, self).__init__()
        self.reduction = reduction
        self.ignore_index = ignore_index

    def forward(self, input, target):
        if self.ignore_index:
            input = input[:,:self.ignore_index,...]
            target = target[:, :self.ignore_index, ...]
        loss = boundary_loss(input, target)
        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        elif self.reduction == 'none':
            return loss

class LabelSmoothingCrossEntropy(torch.nn.Module):
    def __init__(self, smoothing=0.1):
        super().__init__()
        self.smoothing = smoothing

    def forward(self, pred, target):
        confidence = 1. - self.smoothing
        log_probs = F.log_softmax(pred, dim=-1)
        idx = torch.stack([torch.arange(log_probs.shape[0]), target], dim=1)
        nll_loss = torch.gather(-log_probs, dim=-1, index=idx)
        smooth_loss = torch.mean(-log_probs, dim=-1)
        loss = confidence * nll_loss + self.smoothing * smooth_loss

        return loss.mean()


class Joint2loss(nn.Module):
    """
    联合损失函数, 传入两个损失函数
        >>> criterion1 = FocalLoss()
        >>> criterion2 = DiceLoss()
        >>> joint_loss = Joint2loss(criterion1, criterion2, alpha=0.7, reduction='mean')
        >>> input_tensor = torch.rand((1,2,16,16), dtype=torch.float32)
        >>> target_tensor = torch.randn((1,2,16,16), dtype=torch.float32)
        >>> loss = joint_loss(input_tensor, target_tensor)
        >>> print("Joint Loss:", loss.item())
    """
    def __init__(self, *args, alpha, beta=None, ignore_index=None, reduction='mean'):
        super(Joint2loss, self).__init__()
        self.reduction = reduction
        self.ignore_index = ignore_index
        self.alpha = alpha
        self.beta = beta if beta is not None else 1-self.alpha
        self.criterion_1, self.criterion_2 = args

    def forward(self, input, target):
        if self.ignore_index:
            input = input[:, :self.ignore_index,...]
            target = target[:, :self.ignore_index, ...]
        loss_1 = self.criterion_1(input, target)
        loss_2 = self.criterion_2(input, target)
        loss = self.alpha * loss_1 + self.beta * loss_2
        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        elif self.reduction == 'none':
            return loss

if __name__=="__main__":
    criterion1 = FocalLoss()
    criterion2 = DiceLoss()
    joint_loss = Joint2loss(criterion1, criterion2, alpha=0.7, reduction='mean')
    input_tensor = torch.rand((4,2,16,16), dtype=torch.float32)
    target_tensor = torch.randn((4,2,16,16), dtype=torch.float32)
    loss = joint_loss(input_tensor, target_tensor)
    print("Joint Loss:", loss.item())


"""
Copyright (c) 2024, Auorui.
All rights reserved.
用于 分类 的指标检测
                 | Predicted Positive | Predicted Negative |
Actual Positive  |        TP          |         FN         |
Actual Negative  |        FP          |         TN         |
Time: 2024-01-27
"""
import numpy as np
import torch

__all__ = ["accuracy_all_classes", "cls_matrix", "BinaryConfusionMatrix", "MulticlassConfusionMatrix",
           "ConfusionMatrixs", "ModelIndex", "calculate_metrics", "MultiLabelConfusionMatrix"]

def accuracy_all_classes(output, target):
    """
    仅用于所有分类的准确率计算,这里是每个批次的准确率,所以在使用时还要进行itera的累计
    """
    batch_size = target.size(0)
    _, pred = output.max(1)
    accuracy = pred.eq(target).float().sum().item() * 100. / batch_size
    return accuracy

class cls_matrix(object):
    def __init__(self, num_classes):
        self.num_classes = num_classes
        self.matrix = np.zeros((self.num_classes, self.num_classes))
        if self.num_classes <= 1:
            raise ValueError("Argument num_classes needs to be > 1")

    def update(self, pred, true):
        # 在派生类中实现逻辑
        pass

    def reset(self):
        self.matrix = np.zeros((self.num_classes, self.num_classes))

    @property
    def get_matrix(self):
        return self.matrix


class BinaryConfusionMatrix(cls_matrix):
    """
    二分类混淆矩阵类, num_classes默认就为2
    Example:
        >>> confusion_matrix = BinaryConfusionMatrix()
        >>> pred = torch.tensor([0.4, 0.5, 0.5, 0.7, 1, 0.8])
        >>> true = torch.tensor([0, 1, 1, 0, 0, 1])
        >>> confusion_matrix.update(pred, true)
        >>> matrix = confusion_matrix.get_matrix
        >>> print(matrix, confusion_matrix.ravel())
    """
    def __init__(self, threshold=.5):
        super().__init__(num_classes=2)
        self.threshold = threshold
        self.num_classes = 2
        self.reset()

    def update(self, pred, true):
        if torch.is_tensor(pred):
            pred = pred.cpu().detach().numpy()
        if torch.is_tensor(true):
            true = true.cpu().detach().numpy()

        pred = (pred >= self.threshold).astype(int)

        true_positive = np.sum((pred == 1) & (true == 1))
        false_positive = np.sum((pred == 1) & (true == 0))
        true_negative = np.sum((pred == 0) & (true == 0))
        false_negative = np.sum((pred == 0) & (true == 1))

        self.matrix[0, 0] += true_negative
        self.matrix[0, 1] += false_positive
        self.matrix[1, 0] += false_negative
        self.matrix[1, 1] += true_positive

    def ravel(self):
        TP, FN, FP, TN = self.matrix.flatten()
        return TP, FN, FP, TN

class MulticlassConfusionMatrix(cls_matrix):
    """
    多分类混淆矩阵类, num_classes大于2
    Example:
        >>> confusion_matrix = MulticlassConfusionMatrix(num_classes=5, reduction='mean')
        >>> pred = torch.tensor([0, 1, 2, 3, 4, 0, 1, 2, 3, 4])
        >>> true = torch.tensor([0, 1, 2, 3, 4, 1, 2, 3, 4, 0])
        >>> confusion_matrix.update(pred, true)
        >>> matrix = confusion_matrix.get_matrix
        >>> print(matrix, confusion_matrix.ravel())
    """
    def __init__(self, num_classes, reduction='mean'):
        super().__init__(num_classes=num_classes)
        self.reduction = reduction
        self.reset()
        self.threshold = .5

    def update(self, pred, true):
        if torch.is_tensor(pred):
            pred = pred.cpu().detach().numpy()
        if torch.is_tensor(true):
            true = true.cpu().detach().numpy()

        for i in range(self.num_classes):
            for j in range(self.num_classes):
                self.matrix[i, j] += np.sum((true == i) & (pred == j))

    def ravel(self):
        """
        计算混淆矩阵的TN, FP, FN, TP
        支持二分类和多分类
        """
        h = self.matrix.astype(float)
        TP = np.diag(h)
        FN = np.sum(h, axis=1) - TP
        FP = np.sum(h, axis=0) - TP
        TN = np.sum(h) - (np.sum(h, axis=0) + np.sum(h, axis=1) - TP)
        if self.reduction == 'mean':
            return np.mean(TP), np.mean(FN), np.mean(FP), np.mean(TN)
        elif self.reduction == 'sum':
            return np.sum(TP), np.sum(FN), np.sum(FP), np.sum(TN)
        elif self.reduction == 'none':
            return TP, FN, FP, TN

class ConfusionMatrixs(cls_matrix):
    """
    结合多分类与二分类两种情况, 用法与这二者相同
    """
    def __init__(self, num_classes, reduction='mean'):
        super().__init__(num_classes)
        self.num_classes = num_classes
        self.reduction = reduction
        if self.num_classes == 2:
            self.conf = BinaryConfusionMatrix(threshold=0.5)
        elif self.num_classes > 2:
            self.conf = MulticlassConfusionMatrix(self.num_classes, self.reduction)

    def update(self, pred, true):
        self.conf.update(pred, true)

    def ravel(self):
        TP, FN, FP, TN = self.conf.ravel()
        return TP, FN, FP, TN

    @property
    def get_matrix(self):
        return self.conf.get_matrix

def generate_class_weights(num_classes):
    # 类别权重,每个类别平均分配权重
    total_weight = 1.0
    r = 1 if num_classes < 10 else 2
    class_weights = [round(total_weight / num_classes, r)] * (num_classes - 1)
    class_weights.append(round(total_weight - sum(class_weights), r))
    return np.array(class_weights)

class ModelIndex():
    """For details:https://blog.csdn.net/m0_62919535/article/details/132926719
        >>> true_labels = torch.tensor([0, 1, 2, 0, 1, 2])  # 真实标签
        >>> predicted_labels = torch.tensor([0, 1, 1, 0, 2, 1])  # 预测结果
        >>> conf = MulticlassConfusionMatrix(3)
        >>> conf.update(true_labels, predicted_labels)
        >>> TP, FN, FP, TN = conf.ravel()
        >>> index = ModelIndex(TP, FN, FP, TN)
        >>> print(index)

    reduction mean, sum, none, weighted
    """
    def __init__(self,TP, FN, FP, TN, reduction='mean', headline="ModelIndex", esp=1e-5):
        self.TN = TN
        self.FP = FP
        self.FN = FN
        self.TP = TP
        self.reduction = reduction
        self.headline = headline
        self.e = esp
        if self.reduction == 'weighted' or self.reduction == 'none':
            self.num_classes = self.TP.shape[0]
            self.weight = generate_class_weights(self.num_classes)

    @property
    def Accuracy(self):
        """准确度是模型正确分类的样本数量与总样本数量之比"""
        score = (self.TP + self.TN) / (self.TP + self.TN + self.FP + self.FN + self.e)
        if self.reduction == 'mean':
            return score.mean()
        elif self.reduction == 'sum':
            return score.sum()
        elif self.reduction == 'none':
            return score
        elif self.reduction == 'weighted':
            return (score * self.weight).sum()

    @property
    def Precision(self):
        """精确度衡量了正类别预测的准确性"""
        score = self.TP / (self.TP + self.FP + self.e)
        if self.reduction == 'mean':
            return score.mean()
        elif self.reduction == 'sum':
            return score.sum()
        elif self.reduction == 'none':
            return score
        elif self.reduction == 'weighted':
            return (score * self.weight).sum()

    @property
    def Recall(self):
        """召回率衡量了模型对正类别样本的识别能力"""
        score = self.TP / (self.TP + self.FN + self.e)
        if self.reduction == 'mean':
            return score.mean()
        elif self.reduction == 'sum':
            return score.sum()
        elif self.reduction == 'none':
            return score
        elif self.reduction == 'weighted':
            return (score * self.weight).sum()

    @property
    def IOU(self):
        """表示模型预测的区域与真实区域之间的重叠程度"""
        score = self.TP / (self.TP + self.FP + self.FN + self.e)
        if self.reduction == 'mean':
            return score.mean()
        elif self.reduction == 'sum':
            return score.sum()
        elif self.reduction == 'none':
            return score
        elif self.reduction == 'weighted':
            return (score * self.weight).sum()

    @property
    def F1Score(self):
        """F1分数是精确度和召回率的调和平均数"""
        p = self.Precision
        r = self.Recall
        score = (2 * p * r) / (p + r + self.e)
        if self.reduction == 'mean':
            return score.mean()
        elif self.reduction == 'sum':
            return score.sum()
        elif self.reduction == 'none':
            return score
        elif self.reduction == 'weighted':
            return (score * self.weight).sum()

    @property
    def Specificity(self):
        """特异性是指模型在负类别样本中的识别能力"""
        score = self.TN / (self.TN + self.FP + self.e)
        if self.reduction == 'mean':
            return score.mean()
        elif self.reduction == 'sum':
            return score.sum()
        elif self.reduction == 'none':
            return score
        elif self.reduction == 'weighted':
            return (score * self.weight).sum()

    @property
    def FP_rate(self):
        """False Positive Rate,假阳率是模型将负类别样本错误分类为正类别的比例"""
        score = self.FP / (self.FP + self.TN + self.e)
        if self.reduction == 'mean':
            return score.mean()
        elif self.reduction == 'sum':
            return score.sum()
        elif self.reduction == 'none':
            return score
        elif self.reduction == 'weighted':
            return (score * self.weight).sum()

    @property
    def FN_rate(self):
        """False Negative Rate,假阴率是模型将正类别样本错误分类为负类别的比例"""
        score = self.FN / (self.FN + self.TP + self.e)
        if self.reduction == 'mean':
            return score.mean()
        elif self.reduction == 'sum':
            return score.sum()
        elif self.reduction == 'none':
            return score
        elif self.reduction == 'weighted':
            return (score * self.weight).sum()

    @property
    def Qualityfactor(self):
        """品质因子综合考虑了召回率和特异性"""
        r = self.Recall
        s = self.Specificity
        score = r+s-1
        if self.reduction == 'mean':
            return score.mean()
        elif self.reduction == 'sum':
            return score.sum()
        elif self.reduction == 'none':
            return score
        elif self.reduction == 'weighted':
            return (score * self.weight).sum()

    @property
    def Dice(self):
        """Dice系数更多的是用于分割方面"""
        score = 2 * self.TP / (2 * self.TP + self.FN + self.FP + self.e)
        if self.reduction == 'mean':
            return score.mean()
        elif self.reduction == 'sum':
            return score.sum()
        elif self.reduction == 'none':
            return score
        elif self.reduction == 'weighted':
            return (score * self.weight).sum()

    def __str__(self):
        acc_mean = self.Accuracy
        prec_mean = self.Precision
        recall_mean = self.Recall
        iou_mean = self.IOU
        f1_mean = self.F1Score
        spec_mean = self.Specificity
        fp_rate_mean = self.FP_rate
        fn_rate_mean = self.FN_rate
        qf_mean = self.Qualityfactor
        dice_mean = self.Dice
        output_str = f"\033[94m{self.headline}:\n" \
                     f"\t\033[94mClass\tAccuracy\tPrecision    Recall     IOU  \tF1Score\t    Specificity\t   " \
                     f"FP Rate\t  FN Rate\t  Quality Factor\t  Dice\n"
        if self.reduction == 'mean' or self.reduction == 'sum' or self.reduction == 'weighted':
            output_str += (f"\t\033[94mall  \t{acc_mean:.5f}      {prec_mean:.5f}    {recall_mean:.5f}   "
                    f"{iou_mean:.5f}   {f1_mean:.5f}       {spec_mean:.5f}      {fp_rate_mean:.5f}    "
                    f"{fn_rate_mean:.5f}        {qf_mean:.5f}        {dice_mean:.5f}")

        else:
            for i in range(self.TP.shape[0]):
                output_str += f"\t\033[94m{i}\t    {acc_mean[i]:.5f}\t     {prec_mean[i]:.5f}\t{recall_mean[i]:.5f}\t  " \
                              f"{iou_mean[i]:.5f}\t{f1_mean[i]:.5f}\t      {spec_mean[i]:.5f}\t   {fp_rate_mean[i]:.5f}\t  " \
                              f"{fn_rate_mean[i]:.5f}\t     {qf_mean[i]:.5f}\t    {dice_mean[i]:.5f}\n"
            output_str += (f"\t\033[94mmean   \t{acc_mean.mean():.5f}      {prec_mean.mean():.5f}    {recall_mean.mean():.5f}   "
             f"{iou_mean.mean():.5f}   {f1_mean.mean():.5f}       {spec_mean.mean():.5f}      {fp_rate_mean.mean():.5f}    "
             f"{fn_rate_mean.mean():.5f}        {qf_mean.mean():.5f}        {dice_mean.mean():.5f}\n")
            output_str += (f"\t\033[94mweight  {(acc_mean* self.weight).sum():.5f}      {(prec_mean* self.weight).sum():.5f}    {(recall_mean* self.weight).sum():.5f}   "
                           f"{(iou_mean * self.weight).sum():.5f}   {(f1_mean* self.weight).sum():.5f}       {(spec_mean* self.weight).sum():.5f}      {(fp_rate_mean* self.weight).sum():.5f}    "
                           f"{(fn_rate_mean* self.weight).sum():.5f}        {(qf_mean* self.weight).sum():.5f}        {(dice_mean* self.weight).sum():.5f}")
        return output_str

    def get_Index(self):
        metrics_data = {
            'Accuracy': self.Accuracy,
            'Precision': self.Precision,
            'Recall': self.Recall,
            'IOU': self.IOU,
            'F1Score': self.F1Score,
            'Specificity': self.Specificity,
            'FP_rate': self.FP_rate,
            'FN_rate': self.FN_rate,
            'Qualityfactor': self.Qualityfactor,
            'Dice': self.Dice,
        }
        return metrics_data

def calculate_metrics(confusion_matrix):
    num_classes = confusion_matrix.shape[0]
    metrics = np.zeros((4,))

    for i in range(num_classes):
        TP = confusion_matrix[i, i]
        FN = np.sum(confusion_matrix[i, :]) - TP
        FP = np.sum(confusion_matrix[:, i]) - TP
        TN = np.sum(confusion_matrix) - (TP + FN + FP)
        metrics[0] += TP
        metrics[1] += FN
        metrics[2] += FP
        metrics[3] += TN
    return metrics

class MultiLabelConfusionMatrix(cls_matrix):
    def __init__(self, num_classes):
        super().__init__(num_classes=num_classes)
        self.matrixs = 0
        self.reset()

    def reset(self):
        self.matrixs = np.zeros((self.num_classes, 2, 2), dtype=np.int64)

    def update(self, pred, true):
        if torch.is_tensor(pred):
            pred = pred.cpu().detach().numpy()
        if torch.is_tensor(true):
            true = true.cpu().detach().numpy()
        if true.ndim == 1:
            tp = true == pred
            tp_sum = np.sum(tp)
        else:
            true_and_pred = true * pred
            tp_sum = np.count_nonzero(true_and_pred, axis=0)
        fp = np.sum(pred, axis=0) - tp_sum
        fn = np.sum(true, axis=0) - tp_sum
        tp = tp_sum
        tn = true.shape[0] - tp - fp - fn
        self.matrixs += np.array([tn, fp, fn, tp]).T.reshape(-1, 2, 2)

    @property
    def get_matrix(self):
        return self.matrixs

if __name__=="__main__":
    confusion_matrix = ConfusionMatrixs(num_classes=3, reduction='none')
    true = torch.tensor([1, 1, 1, 0, 0, 0, 2, 2, 2, 2])
    pred = torch.tensor([1, 0, 0, 0, 2, 1, 0, 0, 2, 2])
    confusion_matrix.update(pred, true)
    TP, FN, FP, TN = confusion_matrix.ravel()
    print(TP, FN, FP, TN)
    print(confusion_matrix.get_matrix)

    index = ModelIndex(TP, FN, FP, TN, reduction='mean')
    print(index)


"""
Copyright (c) 2024, Auorui.
All rights reserved.

论文原址：https://arxiv.org/pdf/1704.04861.pdf
MobileNet_v1 使用的深度可分离卷积
Depthwise Separable Conv
"""
import torch
import torch.nn as nn

class DepthSepConv(nn.Module):
    """
    深度可分卷积: DW卷积 + PW卷积
    dw卷积, 当分组个数等于输入通道数时, 输出矩阵的通道输也变成了输入通道数
    pw卷积, 使用了1x1的卷积核与普通的卷积一样
    """
    def __init__(self, in_channels, out_channels, stride):
        super(DepthSepConv, self).__init__()
        self.depthwise = nn.Conv2d(in_channels, in_channels, kernel_size=3, stride=stride, groups=in_channels, padding=1)
        self.pointwise = nn.Conv2d(in_channels, out_channels, kernel_size=1)
        self.batch_norm1 = nn.BatchNorm2d(in_channels)
        self.batch_norm2 = nn.BatchNorm2d(out_channels)
        self.relu6 = nn.ReLU6(inplace=True)

    def forward(self, x):
        x = self.depthwise(x)
        x = self.batch_norm1(x)
        x = self.relu6(x)
        x = self.pointwise(x)
        x = self.batch_norm2(x)
        x = self.relu6(x)

        return x

if __name__=="__main__":
    input_tensor = torch.randn(1, 3, 32, 32)
    depthconv = DepthSepConv(3, 6, 1)
    output_tensor = depthconv(input_tensor)
    print("输出张量形状:", output_tensor.shape)



"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is based on cv2 implemented transformations
"""
import cv2
import torch
import random
import numpy as np
from pyzjr.augmentation.augments import Centerzoom, horizontal_flip, vertical_flip, adjust_brightness_cv2, \
    adjust_gamma, augment_Hsv, hist_equalize, random_rotation, random_lighting, blur, gaussian_blur,\
    bilateral_blur, median_blur, random_crop, random_resize_crop

from pyzjr.core.error import _check_img_is_opencv, convert_to_tuple
from pyzjr.Math.Numpy import rand

__all__ = ["OpencvToTensor",
           "OpencvResize",
           "OpencvSquareResize",
           "OpencvCenterzoom",
           "OpencvHorizontalFlip",
           "OpencvVerticalFlip",
           "OpencvBrightness",
           "OpencvAdjustGamma",
           "OpencvToHSV",
           "OpencvHistEqualize",
           "OpencvRotation",
           "OpencvLighting",
           "OpencvRandomBlur",
           "OpencvCrop",
           "OpencvResizeCrop",
           "OpencvPadResize",
           "OpencvGrayscale",
           ]

class OpencvToTensor():
    """OpencvToTensor类用于BGR格式的ToTensor"""
    def __init__(self, half=False):
        super().__init__()
        self.half = half

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        if len(opencv_img.shape) == 3:  # BGR图像
            img = np.ascontiguousarray(opencv_img.transpose((2, 0, 1))[::-1])  # HWC to CHW -> BGR to RGB -> contiguous
        else:  # 灰度图像
            img = np.ascontiguousarray(opencv_img)
        img = torch.from_numpy(img)
        img = img.half() if self.half else img.float()  # uint8 to fp16/32
        img /= 255.0
        return img

class OpencvResize():
    """OpencvResize类用于调整OpenCV图像的大小"""
    def __init__(self, size):
        self.size = convert_to_tuple(size)
        self.interp_method = cv2.INTER_CUBIC

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        image_shape = opencv_img.shape
        if image_shape[-1] > self.size[1]:
            self.interp_method = cv2.INTER_AREA
        image = cv2.resize(opencv_img, self.size, interpolation=self.interp_method)
        return image

class OpencvSquareResize():
    """OpencvSquareResize类用于将图像裁剪为正方形并调整大小"""
    def __init__(self, size):
        super().__init__()
        self.h, self.w = convert_to_tuple(size)

    def __call__(self, opencv_img):  # im = np.array HWC
        _check_img_is_opencv(opencv_img)
        imh, imw = opencv_img.shape[:2]
        m = min(imh, imw)  # min dimension
        top, left = (imh - m) // 2, (imw - m) // 2
        return cv2.resize(opencv_img[top:top + m, left:left + m], (self.w, self.h), interpolation=cv2.INTER_LINEAR)

class OpencvCenterzoom():
    """OpencvCenterzoom类用于中心缩放"""
    def __init__(self, factor):
        super().__init__()
        self.factor = factor

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        return Centerzoom(opencv_img, self.factor)

class OpencvHorizontalFlip():
    """OpencvCenterzoom类用于随机水平翻转"""
    def __init__(self, p=0.5):
        super().__init__()
        self.p = p
        self.random_float = random.random()

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        if self.random_float >= self.p:
            opencv_img = horizontal_flip(opencv_img)
        return opencv_img

class OpencvVerticalFlip():
    """OpencvCenterzoom类用于随机垂直翻转"""
    def __init__(self, p=0.5):
        super().__init__()
        self.p = p
        self.random_float = random.random()

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        if self.random_float >= self.p:
            opencv_img = vertical_flip(opencv_img)
        return opencv_img

class OpencvBrightness():
    """OpencvBrightness类用于亮度调整"""
    def __init__(self, factor):
        super().__init__()
        self.factor = factor

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        opencv_img = adjust_brightness_cv2(opencv_img, self.factor)
        return opencv_img

class OpencvAdjustGamma():
    """OpencvAdjustGamma类用于调整图像的gamma值"""
    def __init__(self, gamma):
        self.gamma = gamma

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        opencv_img = adjust_gamma(opencv_img, self.gamma)
        return opencv_img


class OpencvToHSV():
    """OpencvToHSV类用于转换图像到HSV颜色空间"""
    def __init__(self, hue_range=(-0.1, 0.3), saturation_range=(0.5, 1.4), value_range=(0.7, 1.3)):
        self.hgain = rand(hue_range[0], hue_range[1])
        self.sgain = rand(saturation_range[0], saturation_range[1])
        self.vgain = rand(value_range[0], value_range[1])

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        opencv_img = augment_Hsv(opencv_img, self.hgain, self.sgain, self.vgain)
        return opencv_img

class OpencvHistEqualize():
    """OpencvHistEqualize类用于对图像进行直方图均衡化"""
    def __init__(self, clahe=True):
        self.clahe = clahe

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        opencv_img = hist_equalize(opencv_img, self.clahe)
        return opencv_img

class OpencvRotation():
    """OpencvRotation类用于对图像进行随机旋转"""
    def __init__(self, degrees=None):
        self.degrees = degrees

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        opencv_img = random_rotation(opencv_img, self.degrees)
        return opencv_img

class OpencvLighting():
    """OpencvLighting类用于对图像进行随机光照调整"""
    def __init__(self, alpha):
        self.alpha = alpha

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        opencv_img = random_lighting(opencv_img, self.alpha)
        return opencv_img

class OpencvRandomBlur():
    """OpencvRandomBlur类用于对图像进行随机模糊(支持四选一)"""
    def __init__(self, ksize, choice=None):
        self.ksize = ksize
        self.choice = choice
        if self.ksize not in {1, 3, 5, 7, 9}:
            raise ValueError(f"Invalid ksize value {ksize}.The available values are 1, 3, 5, 7, 9")
        if self.choice is None:
            filter_functions = ["blur", "median", "gaussian", "bilateral"]
            self.choice = random.choice(filter_functions)

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        if self.choice == "blur":
            self.ksize = convert_to_tuple(self.ksize)
            opencv_img = blur(opencv_img, self.ksize)
        elif self.choice == "median":
            opencv_img = median_blur(opencv_img, self.ksize)
        elif self.choice == "gaussian":
            opencv_img = gaussian_blur(opencv_img, ksize=self.ksize)
        elif self.choice == "bilateral":
            opencv_img = bilateral_blur(opencv_img, self.ksize)
        return opencv_img

class OpencvCrop():
    """OpencvCrop类用于对图像进行随机裁剪"""
    def __init__(self, size):
        self.size = convert_to_tuple(size)

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        height, width = opencv_img.shape[:2]
        if self.size[0] > height or self.size[1] > width:
            raise ValueError("The cropped size cannot exceed the original image size")
        opencv_img = random_crop(opencv_img, self.size)
        return opencv_img

class OpencvResizeCrop():
    """OpencvResizeCrop类用于对图像进行随机缩放和裁剪"""
    def __init__(self, size, scale_range=(1.5, 2.5)):
        self.size = convert_to_tuple(size)
        self.scale_range = scale_range

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        opencv_img = random_resize_crop(opencv_img, self.size, self.scale_range)
        return opencv_img

class OpencvPadResize():
    """OpencvPadResize类用于调整OpenCV图像的大小并在背景中居中"""
    def __init__(self, size):
        self.size = convert_to_tuple(size)
        self.full = (128, 128, 128)

    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        height, width = opencv_img.shape[:2]
        h, w = self.size
        scale = min(w / width, h / height)
        nw = int(width * scale)
        nh = int(height * scale)
        opencv_img = cv2.resize(opencv_img, (nw, nh))
        background = np.full((h, w, 3), self.full, dtype=np.uint8)
        x_offset = (w - nw) // 2
        y_offset = (h - nh) // 2
        background[y_offset:y_offset+nh, x_offset:x_offset+nw] = opencv_img

        return background

class OpencvGrayscale():
    """OpencvGrayscale类用于将BGR图像转为灰度图再转回BGR,一般会导致颜色信息丢失"""
    def __call__(self, opencv_img):
        _check_img_is_opencv(opencv_img)
        opencv_img = cv2.cvtColor(opencv_img, cv2.COLOR_BGR2GRAY)
        opencv_img = cv2.cvtColor(opencv_img, cv2.COLOR_GRAY2BGR)
        return opencv_img

if __name__=="__main__":
    from pyzjr.augmentation.transforms._utils import Compose
    img = cv2.imread(r"D:\PythonProject\pyzjrPyPi\pyzjr\augmentation\test.png")
    transforms = Compose([
        OpencvCenterzoom(1.1),
        OpencvResize(256),
        # OpencvGrayscale(),
        OpencvBrightness(1.4),
        OpencvHorizontalFlip(p=0.5),
        OpencvVerticalFlip(p=0.5),
        OpencvHistEqualize(),
        OpencvAdjustGamma(1.2),
        OpencvLighting(1.2),
        OpencvRotation(),
        OpencvRandomBlur(3, choice="gaussian"),
        OpencvPadResize(512),
        OpencvCrop(300),
        OpencvResizeCrop(200),
        OpencvToTensor(),
    ])
    transformed_image = transforms(img)
    print(transformed_image.shape)
    # 将RGB转为BGR
    transformed_image_bgr = torch.flip(transformed_image, dims=[0])
    img = (transformed_image.permute((1, 2, 0)).numpy() * 255).astype(np.uint8)
    cv2.imshow("ss", img)
    cv2.waitKey(0)







"""
compute_sdf1_1、compute_sdf、boundary_loss
Reference from https://github.com/LIVIAETS/boundary-loss?tab=readme-ov-file
time 2024-01-26
"""
import numpy as np
import torch
from scipy.ndimage import distance_transform_edt
from skimage import segmentation as skimage_seg

def compute_sdf1_1(img_gt, out_shape):
    """
    compute the normalized signed distance map of binary mask
    input: segmentation, shape = (batch_size, c, x, y)
    output: the Signed Distance Map (SDM)
    sdf(x) = 0; x in segmentation boundary
             -inf|x-y|; x in segmentation
             +inf|x-y|; x out of segmentation
    normalize sdf to [-1, 1]
    """
    img_gt = img_gt.astype(np.uint8)
    normalized_sdf = np.zeros(out_shape)

    for b in range(out_shape[0]):  # batch size
        # ignore background
        posmask = img_gt[b].astype(bool)
        for c in range(out_shape[1]):
            if posmask.any():
                negmask = ~posmask
                posdis = distance_transform_edt(posmask)
                negdis = distance_transform_edt(negmask)
                boundary = skimage_seg.find_boundaries(posmask, mode='inner').astype(np.uint8)
                sdf = (negdis - np.min(negdis)) / (np.max(negdis) - np.min(negdis)) - (posdis - np.min(posdis)) / (
                            np.max(posdis) - np.min(posdis))
                sdf[boundary == 1] = 0
                normalized_sdf[b][c] = sdf

    return normalized_sdf


def compute_sdf(img_gt, out_shape):
    """
    compute the signed distance map of binary mask
    input: segmentation, shape = (batch_size, x, y, z)
    output: the Signed Distance Map (SDM)
    sdf(x) = 0; x in segmentation boundary
             -inf|x-y|; x in segmentation
             +inf|x-y|; x out of segmentation
    """
    img_gt = img_gt.astype(np.uint8)
    gt_sdf = np.zeros(out_shape)

    for b in range(out_shape[0]):  # batch size
        posmask = img_gt[b].astype(bool)
        for c in range(out_shape[1]):
            if posmask.any():
                negmask = ~posmask
                posdis = distance_transform_edt(posmask)
                negdis = distance_transform_edt(negmask)
                boundary = skimage_seg.find_boundaries(posmask, mode='inner').astype(np.uint8)
                sdf = negdis - posdis
                sdf[boundary == 1] = 0
                gt_sdf[b][c] = sdf

    return gt_sdf

def boundary_loss(outputs_soft, label_batch):
    """
    compute boundary loss for binary segmentation
    Args:
        outputs_soft: sigmoid results,  shape=(b,c,x,y)
        label_batch: sdf of ground truth (can be original or normalized sdf); shape=(b,c,x,y)

    Returns: boundary_loss
    """
    gt_sdf_npy = compute_sdf1_1(label_batch.numpy(), outputs_soft.shape)
    gt_sdf = torch.from_numpy(gt_sdf_npy).float()
    pc = outputs_soft.type(torch.float32)
    dc = gt_sdf.type(torch.float32)

    multipled = torch.einsum("bcwh, bcwh->bcwh", pc, dc)
    loss = multipled.mean(dim=(1, 2, 3))

    return loss


if __name__=="__main__":
    batch_size, channels, height, width = 1, 1, 3, 3
    # model_predictions = torch.rand((batch_size, channels, height, width))
    model_predictions = torch.Tensor([[[[0.7616, 0.0239, 0.5281],
                                        [0.3337, 0.5539, 0.3333],
                                        [0.1318, 0.8004, 0.5439]]]])

    model_predictions = torch.sigmoid(model_predictions)
    seg_mask = np.random.randint(2, size=(batch_size, channels, height, width))
    seg_mask_tensor = torch.from_numpy(seg_mask).float()
    loss = boundary_loss(model_predictions, seg_mask_tensor)

    # Print the loss value
    print("Boundary Loss:", loss.item())



"""
`MnasNet: Platform-Aware Neural Architecture Search for Mobile
<https://arxiv.org/pdf/1807.11626.pdf>`_ paper.
"""
import torch
import torch.nn as nn
from functools import partial

__all__ = ["MNASNet", "mnasnet0_5", "mnasnet0_75", "mnasnet1_0", "mnasnet1_3",]


# Paper suggests 0.9997 momentum, for TensorFlow. Equivalent PyTorch momentum is 1.0 - tensorflow.
_BN_MOMENTUM = 1 - 0.9997
BatchNorm2d = partial(nn.BatchNorm2d, eps=0.001, momentum=_BN_MOMENTUM)

class _InvertedResidual(nn.Module):
    def __init__(self, in_ch, out_ch, kernel_size, stride, expansion_factor, bn_momentum=0.1):
        super(_InvertedResidual, self).__init__()
        if stride not in [1, 2]:
            raise ValueError(f"stride should be 1 or 2 instead of {stride}")
        if kernel_size not in [3, 5]:
            raise ValueError(f"kernel_size should be 3 or 5 instead of {kernel_size}")
        mid_ch = in_ch * expansion_factor
        self.apply_residual = in_ch == out_ch and stride == 1
        self.layers = nn.Sequential(
            # Pointwise
            nn.Conv2d(in_ch, mid_ch, 1, bias=False),
            nn.BatchNorm2d(mid_ch, momentum=bn_momentum),
            nn.ReLU(inplace=True),
            # Depthwise
            nn.Conv2d(mid_ch, mid_ch, kernel_size, padding=kernel_size // 2, stride=stride, groups=mid_ch, bias=False),
            nn.BatchNorm2d(mid_ch, momentum=bn_momentum),
            nn.ReLU(inplace=True),
            # Linear pointwise. Note that there's no activation.
            nn.Conv2d(mid_ch, out_ch, 1, bias=False),
            nn.BatchNorm2d(out_ch, momentum=bn_momentum),
        )

    def forward(self, input):
        if self.apply_residual:
            return self.layers(input) + input
        else:
            return self.layers(input)


def _stack(in_ch, out_ch, kernel_size, stride, exp_factor, repeats, bn_momentum):
    """Creates a stack of inverted residuals."""
    if repeats < 1:
        raise ValueError(f"repeats should be >= 1, instead got {repeats}")
    # First one has no skip, because feature map size changes.
    first = _InvertedResidual(in_ch, out_ch, kernel_size, stride, exp_factor, bn_momentum=bn_momentum)
    remaining = []
    for _ in range(1, repeats):
        remaining.append(_InvertedResidual(out_ch, out_ch, kernel_size, 1, exp_factor, bn_momentum=bn_momentum))
    return nn.Sequential(first, *remaining)

def _make_divisible(v, divisor, min_value=None):
    """
    This function is taken from the original tf repo.It can be seen here:
        https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    Args:
        v: The number of input channels.
        divisor: The number of channels should be a multiple of this value.
        min_value: The minimum value of the number of channels, which defaults to the advisor.

    Returns: It ensures that all layers have a channel number that is divisible by 8
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v

def _get_depths(alpha):
    """Scales tensor depths as in reference MobileNet code, prefers rouding up
    rather than down."""
    depths = [32, 16, 24, 40, 80, 96, 192, 320]
    return [_make_divisible(depth * alpha, 8) for depth in depths]


class MNASNet(nn.Module):
    def __init__(self, alpha, num_classes=1000, drop_rate=.2):
        super().__init__()
        if alpha <= 0.0:
            raise ValueError(f"alpha should be greater than 0.0 instead of {alpha}")
        self.alpha = alpha
        self.num_classes = num_classes
        depths = _get_depths(alpha)
        # First layer
        self.SepConv = nn.Sequential(
            # Conv3x3
            nn.Conv2d(3, depths[0], kernel_size=3, padding=1, stride=2, bias=False),   # torch.Size([2, 32, 112, 112])
            BatchNorm2d(depths[0]),
            nn.ReLU(inplace=True),
            # Depthwise separable
            nn.Conv2d(depths[0], depths[0], 3, padding=1, stride=1, groups=depths[0], bias=False),  # torch.Size([2, 32, 112, 112])
            BatchNorm2d(depths[0]),
            nn.ReLU(inplace=True),
            # Conv1x1
            nn.Conv2d(depths[0], depths[1], 1, padding=0, stride=1, bias=False),  # torch.Size([2, 16, 112, 112])
            BatchNorm2d(depths[1]),
        )
        self.inverted_residuals = nn.Sequential(
            # MNASNet blocks: stacks of inverted residuals.
            _stack(depths[1], depths[2], 3, 2, 3, 3, _BN_MOMENTUM),
            _stack(depths[2], depths[3], 5, 2, 3, 3, _BN_MOMENTUM),
            _stack(depths[3], depths[4], 5, 2, 6, 3, _BN_MOMENTUM),
            _stack(depths[4], depths[5], 3, 1, 6, 2, _BN_MOMENTUM),
            _stack(depths[5], depths[6], 5, 2, 6, 4, _BN_MOMENTUM),
            _stack(depths[6], depths[7], 3, 1, 6, 1, _BN_MOMENTUM),
            # Final mapping to classifier input.
            nn.Conv2d(depths[7], 1280, 1, padding=0, stride=1, bias=False),
            BatchNorm2d(1280, momentum=_BN_MOMENTUM),
            nn.ReLU(inplace=True),
        )
        self.classifier = nn.Sequential(
            nn.Dropout(p=drop_rate, inplace=True),
            nn.Linear(1280, num_classes)
        )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.kaiming_uniform_(m.weight, mode="fan_out", nonlinearity="sigmoid")
                nn.init.zeros_(m.bias)

    def forward(self, x):
        x = self.SepConv(x)
        x = self.inverted_residuals(x)
        # Equivalent to global avgpool and removing H and W dimensions.
        x = x.mean([2, 3])
        x = self.classifier(x)
        return x


def mnasnet0_5(num_classes, **kwargs):
    return MNASNet(0.5, num_classes=num_classes, **kwargs)


def mnasnet0_75(num_classes, **kwargs):
    return MNASNet(0.75, num_classes=num_classes, **kwargs)


def mnasnet1_0(num_classes, **kwargs):
    return MNASNet(1.0, num_classes=num_classes, **kwargs)


def mnasnet1_3(num_classes, **kwargs):
    return MNASNet(1.3, num_classes=num_classes, **kwargs)

if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = mnasnet1_0(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # Total params: 5,006,380


# 像素级别的研究

from .pixel import SkeletonMap, incircle, outcircle
from .Crack import infertype, CrackType, DetectCracks
from ._Steger import Steger
from .utils import DistanceTransform
from .definition import *

__all__ = ["SkeletonMap", "incircle", "outcircle", "infertype", "CrackType", "DetectCracks",
           "Steger", "DistanceTransform"]


"""
Creates a GhostNet Model as defined in:
GhostNet: More Features from Cheap Operations By Kai Han, Yunhe Wang, Qi Tian, Jianyuan Guo, Chunjing Xu, Chang Xu.
ghostnetv1: <https://arxiv.org/pdf/1911.11907.pdf>
ghostnetv2: <https://arxiv.org/pdf/2211.12905v1.pdf>
g_ghostnet: <https://arxiv.org/pdf/2201.03297.pdf>
"""
import math
import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ["GhostNet", "GGhostRegNet", "ghostnetv1", "ghostnetv2", "g_ghost_regnetx_002", "g_ghost_regnetx_004",
           "g_ghost_regnetx_006", "g_ghost_regnetx_008", "g_ghost_regnetx_016", "g_ghost_regnetx_032",
           "g_ghost_regnetx_040", "g_ghost_regnetx_064", "g_ghost_regnetx_080", "g_ghost_regnetx_120",
           "g_ghost_regnetx_160", "g_ghost_regnetx_320"]

def _make_divisible(v, divisor, min_value=None):
    """
    This function is taken from the original tf repo.It can be seen here:
        https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    Args:
        v: The number of input channels.
        divisor: The number of channels should be a multiple of this value.
        min_value: The minimum value of the number of channels, which defaults to the advisor.

    Returns: It ensures that all layers have a channel number that is divisible by 8
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class SeModule(nn.Module):
    def __init__(self, input_channels, se_ratio=0.25, divisor=4):
        super(SeModule, self).__init__()
        reduced_chs = _make_divisible(input_channels * se_ratio, divisor)
        self.se = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(input_channels, reduced_chs, kernel_size=1, bias=True),
            nn.BatchNorm2d(reduced_chs),
            nn.ReLU(inplace=True),
            nn.Conv2d(reduced_chs, input_channels, kernel_size=1, bias=True),
            nn.Hardsigmoid()
        )

    def forward(self, x):
        x = x * self.se(x)
        return x

class GhostModulev1(nn.Module):
    def __init__(self, inp, oup, kernel_size=1, ratio=2, dw_size=3, stride=1, relu=True):
        super(GhostModulev1, self).__init__()
        self.oup = oup
        init_channels = math.ceil(oup / ratio)   # m = n / s
        new_channels = init_channels*(ratio-1)   # m * (s - 1) = n / s * (s - 1)

        # 利用1x1卷积对输入进来的特征图进行通道的浓缩, 实现跨通道的特征提取
        self.primary_conv = nn.Sequential(
            nn.Conv2d(inp, init_channels, kernel_size, stride, kernel_size//2, bias=False),
            nn.BatchNorm2d(init_channels),
            nn.ReLU(inplace=True) if relu else nn.Sequential(),
        )
        # 使用逐层卷积, 获得额外的特征图
        self.cheap_operation = nn.Sequential(
            nn.Conv2d(init_channels, new_channels, dw_size, stride=1, padding=dw_size//2, groups=init_channels, bias=False),
            nn.BatchNorm2d(new_channels),
            nn.ReLU(inplace=True) if relu else nn.Sequential(),
        )

    def forward(self, x):
        x1 = self.primary_conv(x)
        x2 = self.cheap_operation(x1)
        out = torch.cat([x1, x2], dim=1)
        return out[:, :self.oup, :, :]


class GhostBottleneckv1(nn.Module):
    """ Ghost bottleneck w/ optional SE"""
    def __init__(self, in_chs, mid_chs, out_chs, dw_size=3, stride=1, se_ratio=0.):
        super(GhostBottleneckv1, self).__init__()
        has_se = se_ratio is not None and se_ratio > 0.
        self.stride = stride

        # Point-wise expansion
        self.ghost1 = GhostModulev1(in_chs, mid_chs, relu=True)

        # Depth-wise convolution
        if self.stride > 1:
            self.conv_dw = nn.Conv2d(mid_chs, mid_chs, dw_size, stride=stride,
                                     padding=(dw_size-1)//2,
                                     groups=mid_chs, bias=False)
            self.bn_dw = nn.BatchNorm2d(mid_chs)

        # Squeeze-and-excitation
        if has_se:
            self.se = SeModule(mid_chs, se_ratio=se_ratio)
        else:
            self.se = None

        # Point-wise linear projection
        self.ghost2 = GhostModulev1(mid_chs, out_chs, relu=False)

        # shortcut
        if (in_chs == out_chs and self.stride == 1):
            self.shortcut = nn.Sequential()
        else:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_chs, in_chs, dw_size, stride=stride,
                          padding=(dw_size-1)//2, groups=in_chs, bias=False),
                nn.BatchNorm2d(in_chs),
                nn.Conv2d(in_chs, out_chs, 1, stride=1, padding=0, bias=False),
                nn.BatchNorm2d(out_chs),
            )

    def forward(self, x):
        residual = x
        # 1st ghost bottleneck
        x = self.ghost1(x)
        # Depth-wise convolution
        if self.stride > 1:
            x = self.conv_dw(x)
            x = self.bn_dw(x)
        # Squeeze-and-excitation
        if self.se is not None:
            x = self.se(x)
        # 2nd ghost bottleneck
        x = self.ghost2(x)
        x += self.shortcut(residual)
        return x


class GhostModulev2(nn.Module):
    """添加了DFC注意力机制,即short_conv部分,原文使用的是max_pool,这里使用的是avg_pool"""
    def __init__(self, inp, oup, kernel_size=1, ratio=2, dw_size=3, stride=1, relu=True, mode=None):
        super(GhostModulev2, self).__init__()
        self.mode = mode
        self.gate_fn = nn.Sigmoid()

        if self.mode in ['original']:
            self.oup = oup
            init_channels = math.ceil(oup / ratio)
            new_channels = init_channels*(ratio-1)
            self.primary_conv = nn.Sequential(
                nn.Conv2d(inp, init_channels, kernel_size, stride, kernel_size//2, bias=False),
                nn.BatchNorm2d(init_channels),
                nn.ReLU(inplace=True) if relu else nn.Sequential(),
            )
            self.cheap_operation = nn.Sequential(
                nn.Conv2d(init_channels, new_channels, dw_size, 1, dw_size//2, groups=init_channels, bias=False),
                nn.BatchNorm2d(new_channels),
                nn.ReLU(inplace=True) if relu else nn.Sequential(),
            )
        elif self.mode in ['attn']:
            self.oup = oup
            init_channels = math.ceil(oup / ratio)
            new_channels = init_channels*(ratio-1)
            self.primary_conv = nn.Sequential(
                nn.Conv2d(inp, init_channels, kernel_size, stride, kernel_size//2, bias=False),
                nn.BatchNorm2d(init_channels),
                nn.ReLU(inplace=True) if relu else nn.Sequential(),
            )
            self.cheap_operation = nn.Sequential(
                nn.Conv2d(init_channels, new_channels, dw_size, 1, dw_size//2, groups=init_channels, bias=False),
                nn.BatchNorm2d(new_channels),
                nn.ReLU(inplace=True) if relu else nn.Sequential(),
            )
            self.short_conv = nn.Sequential(
                # horizontal DFC and vertical DFC
                nn.Conv2d(inp, oup, kernel_size, stride, kernel_size//2, bias=False),
                nn.BatchNorm2d(oup),
                nn.Conv2d(oup, oup, kernel_size=(1, 5), stride=1, padding=(0,2), groups=oup,bias=False),
                nn.BatchNorm2d(oup),
                nn.Conv2d(oup, oup, kernel_size=(5, 1), stride=1, padding=(2,0), groups=oup,bias=False),
                nn.BatchNorm2d(oup),
            )

    def forward(self, x):
        if self.mode in ['original']:
            x1 = self.primary_conv(x)
            x2 = self.cheap_operation(x1)
            out = torch.cat([x1, x2], dim=1)
            return out[:, : self.oup, :, :]
        elif self.mode in ['attn']:
            res = self.short_conv(F.avg_pool2d(x, kernel_size=2, stride=2))
            x1 = self.primary_conv(x)
            x2 = self.cheap_operation(x1)
            out = torch.cat([x1, x2], dim=1)
            return out[:, :self.oup, :, :] * F.interpolate(self.gate_fn(res), size=(out.shape[-2], out.shape[-1]), mode='nearest')

class GhostBottleneckv2(nn.Module):
    """与原文有所出入,在消融实验中每层都添加了DFC,这里前两层还是用的原本v1版本"""
    def __init__(self, in_chs, mid_chs, out_chs, dw_kernel_size=3,
                 stride=1, se_ratio=0., layer_id=None):
        super(GhostBottleneckv2, self).__init__()
        has_se = se_ratio is not None and se_ratio > 0.
        self.stride = stride

        # Point-wise expansion
        if layer_id <= 1:
            self.ghost1 = GhostModulev2(in_chs, mid_chs, relu=True, mode='original')
        else:
            self.ghost1 = GhostModulev2(in_chs, mid_chs, relu=True, mode='attn')

        # Depth-wise convolution
        if self.stride > 1:
            self.conv_dw = nn.Conv2d(mid_chs, mid_chs, dw_kernel_size, stride=stride,
                                     padding=(dw_kernel_size-1)//2, groups=mid_chs, bias=False)
            self.bn_dw = nn.BatchNorm2d(mid_chs)

        # Squeeze-and-excitation
        if has_se:
            self.se = SeModule(mid_chs, se_ratio=se_ratio)
        else:
            self.se = None

        self.ghost2 = GhostModulev2(mid_chs, out_chs, relu=False,mode='original')

        # shortcut
        if (in_chs == out_chs and self.stride == 1):
            self.shortcut = nn.Sequential()
        else:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_chs, in_chs, dw_kernel_size, stride=stride,
                          padding=(dw_kernel_size-1)//2, groups=in_chs, bias=False),
                nn.BatchNorm2d(in_chs),
                nn.Conv2d(in_chs, out_chs, 1, stride=1, padding=0, bias=False),
                nn.BatchNorm2d(out_chs),
            )

    def forward(self, x):
        residual = x
        x = self.ghost1(x)
        if self.stride > 1:
            x = self.conv_dw(x)
            x = self.bn_dw(x)
        if self.se is not None:
            x = self.se(x)
        x = self.ghost2(x)
        x += self.shortcut(residual)
        return x

class GhostNet(nn.Module):
    def __init__(self, cfgs, num_classes=1000, width=1.0, drop_rate=0.2, block=None):
        super(GhostNet, self).__init__()
        self.cfgs = cfgs
        self.drop_rate = drop_rate

        # building first layer
        output_channel = _make_divisible(16 * width, 4)
        self.conv_stem = nn.Conv2d(3, output_channel, 3, 2, 1, bias=False)
        # 224 x 224 x 3 --> 112 x 112 x 16
        self.bn1 = nn.BatchNorm2d(output_channel)
        self.act1 = nn.ReLU(inplace=True)
        input_channel = output_channel

        # building inverted residual blocks
        stages = []
        layer_id = 0
        for cfg in self.cfgs:
            layers = []
            for k, exp_size, c, se_ratio, s in cfg:
                output_channel = _make_divisible(c * width, 4)
                hidden_channel = _make_divisible(exp_size * width, 4)
                if block == GhostBottleneckv2:
                    layers.append(block(input_channel, hidden_channel, output_channel, k, s,
                                        se_ratio=se_ratio, layer_id=layer_id))
                elif block == GhostBottleneckv1:
                    layers.append(block(input_channel, hidden_channel, output_channel, k, s,
                                        se_ratio=se_ratio))
                input_channel = output_channel
                layer_id += 1
            stages.append(nn.Sequential(*layers))

        output_channel = _make_divisible(exp_size * width, 4)
        stages.append(nn.Sequential(
            nn.Conv2d(input_channel, output_channel, kernel_size=1, stride=1, padding='same', bias=False),
            nn.BatchNorm2d(output_channel),
            nn.ReLU(inplace=True),
        ))
        input_channel = output_channel

        self.blocks = nn.Sequential(*stages)

        # building last several layers
        output_channel = 1280
        self.global_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.conv_head = nn.Conv2d(input_channel, output_channel, 1, 1, 0, bias=True)
        self.act2 = nn.ReLU(inplace=True)
        self.classifier = nn.Linear(output_channel, num_classes)

    def forward(self, x):
        x = self.conv_stem(x)
        x = self.bn1(x)
        x = self.act1(x)
        x = self.blocks(x)
        x = self.global_pool(x)
        x = self.conv_head(x)
        x = self.act2(x)
        x = x.view(x.size(0), -1)
        if self.drop_rate > 0.:
            x = F.dropout(x, p=self.drop_rate, training=self.training)
        x = self.classifier(x)
        return x


def conv3x3(in_channels, out_channels, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding:捕捉局部特征和空间相关性，学习更复杂的特征和抽象表示"""
    return nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)

def conv1x1(in_channels, out_channels, stride=1):
    """1x1 convolution:实现降维或升维，调整通道数和执行通道间的线性变换"""
    return nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)


class Bottleneck(nn.Module):
    expansion = 1
    def __init__(self, inplanes, planes, stride=1, downsample=None, group_width=1, dilation=1):
        super(Bottleneck, self).__init__()
        width = planes * self.expansion
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = nn.BatchNorm2d(width)
        self.conv2 = conv3x3(width, width, stride, width // min(width, group_width), dilation)
        self.bn2 = nn.BatchNorm2d(width)
        self.conv3 = conv1x1(width, planes)
        self.bn3 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class LambdaLayer(nn.Module):
    def __init__(self, lambd):
        super(LambdaLayer, self).__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)


class Stage(nn.Module):
    def __init__(self, block, inplanes, planes, group_width, blocks, stride=1, dilate=False, cheap_ratio=0.5):
        super(Stage, self).__init__()
        downsample = None
        self.dilation = 1
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes:
            downsample = nn.Sequential(
                conv1x1(inplanes, planes, stride),
                nn.BatchNorm2d(planes),
            )

        self.base = block(inplanes, planes, stride, downsample, group_width,
                          previous_dilation)
        self.end = block(planes, planes, group_width=group_width,
                         dilation=self.dilation,)

        group_width = int(group_width * 0.75)
        raw_planes = int(planes * (1 - cheap_ratio) / group_width) * group_width
        cheap_planes = planes - raw_planes
        self.cheap_planes = cheap_planes
        self.raw_planes = raw_planes

        self.merge = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(planes + raw_planes * (blocks - 2), cheap_planes,
                      kernel_size=1, stride=1, bias=False),
            nn.BatchNorm2d(cheap_planes),
            nn.ReLU(inplace=True),
            nn.Conv2d(cheap_planes, cheap_planes, kernel_size=1, bias=False),
            nn.BatchNorm2d(cheap_planes),
        )
        self.cheap = nn.Sequential(
            nn.Conv2d(cheap_planes, cheap_planes,
                      kernel_size=1, stride=1, bias=False),
            nn.BatchNorm2d(cheap_planes),
        )
        self.cheap_relu = nn.ReLU(inplace=True)
        downsample = nn.Sequential(
            LambdaLayer(lambda x: x[:, :raw_planes])
        )

        layers = []
        layers.append(block(raw_planes, raw_planes, 1, downsample, group_width,
                            self.dilation))
        inplanes = raw_planes
        for _ in range(2, blocks - 1):
            layers.append(block(inplanes, raw_planes, group_width=group_width,
                                dilation=self.dilation,))

        self.layers = nn.Sequential(*layers)

    def forward(self, input):
        x0 = self.base(input)

        m_list = [x0]
        e = x0[:, :self.raw_planes]
        for layer in self.layers:
            e = layer(e)
            m_list.append(e)
        m = torch.cat(m_list, 1)
        m = self.merge(m)

        c = x0[:, self.raw_planes:]
        c = self.cheap_relu(self.cheap(c) + m)

        x = torch.cat((e, c), 1)
        x = self.end(x)
        return x


class GGhostRegNet(nn.Module):
    def __init__(self, block, layers, widths, num_classes=1000,
                 group_width=1):
        super(GGhostRegNet, self).__init__()
        self.inplanes = 32
        self.dilation = 1
        replace_stride_with_dilation = [False, False, False, False]
        self.group_width = group_width
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=3, stride=2, padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(self.inplanes)
        self.relu = nn.ReLU(inplace=True)

        self.layer1 = self._make_layer(block, widths[0], layers[0], stride=2,
                                       dilate=replace_stride_with_dilation[0])

        self.inplanes = widths[0]
        if layers[1] > 2:
            self.layer2 = Stage(block, self.inplanes, widths[1], group_width, layers[1], stride=2,
                                dilate=replace_stride_with_dilation[1], cheap_ratio=0.5)
        else:
            self.layer2 = self._make_layer(block, widths[1], layers[1], stride=2,
                                           dilate=replace_stride_with_dilation[1])

        self.inplanes = widths[1]
        self.layer3 = Stage(block, self.inplanes, widths[2], group_width, layers[2], stride=2,
                            dilate=replace_stride_with_dilation[2], cheap_ratio=0.5)

        self.inplanes = widths[2]
        if layers[3] > 2:
            self.layer4 = Stage(block, self.inplanes, widths[3], group_width, layers[3], stride=2,
                                dilate=replace_stride_with_dilation[3], cheap_ratio=0.5)
        else:
            self.layer4 = self._make_layer(block, widths[3], layers[3], stride=2,
                                           dilate=replace_stride_with_dilation[3])
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(0.2)
        self.fc = nn.Linear(widths[-1] * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes, stride),
                nn.BatchNorm2d(planes),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.group_width,
                            previous_dilation))
        self.inplanes = planes
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, group_width=self.group_width,
                                dilation=self.dilation,))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.dropout(x)
        x = self.fc(x)

        return x



def g_ghost_regnetx_002(**kwargs):
    return GGhostRegNet(Bottleneck, [1, 1, 4, 7], [24, 56, 152, 368], group_width=8, **kwargs)


def g_ghost_regnetx_004(**kwargs):
    return GGhostRegNet(Bottleneck, [1, 2, 7, 12], [32, 64, 160, 384], group_width=16, **kwargs)


def g_ghost_regnetx_006(**kwargs):
    return GGhostRegNet(Bottleneck, [1, 3, 5, 7], [48, 96, 240, 528], group_width=24, **kwargs)


def g_ghost_regnetx_008(**kwargs):
    return GGhostRegNet(Bottleneck, [1, 3, 7, 5], [64, 128, 288, 672], group_width=16, **kwargs)


def g_ghost_regnetx_016(**kwargs):
    return GGhostRegNet(Bottleneck, [2, 4, 10, 2], [72, 168, 408, 912], group_width=24, **kwargs)


def g_ghost_regnetx_032(**kwargs):
    return GGhostRegNet(Bottleneck, [2, 6, 15, 2], [96, 192, 432, 1008], group_width=48, **kwargs)


def g_ghost_regnetx_040(**kwargs):
    return GGhostRegNet(Bottleneck, [2, 5, 14, 2], [80, 240, 560, 1360], group_width=40, **kwargs)


def g_ghost_regnetx_064(**kwargs):
    return GGhostRegNet(Bottleneck, [2, 4, 10, 1], [168, 392, 784, 1624], group_width=56, **kwargs)


def g_ghost_regnetx_080(**kwargs):
    return GGhostRegNet(Bottleneck, [2, 5, 15, 1], [80, 240, 720, 1920], group_width=120, **kwargs)


def g_ghost_regnetx_120(**kwargs):
    return GGhostRegNet(Bottleneck, [2, 5, 11, 1], [224, 448, 896, 2240], group_width=112, **kwargs)


def g_ghost_regnetx_160(**kwargs):
    return GGhostRegNet(Bottleneck, [2, 6, 13, 1], [256, 512, 896, 2048], group_width=128, **kwargs)


def g_ghost_regnetx_320(**kwargs):
    return GGhostRegNet(Bottleneck, [2, 7, 13, 1], [336, 672, 1344, 2520], group_width=168, **kwargs)


def ghostnetv1(num_classes):
    """
    * k: 卷积核大小,表示跨特征点提取特征的能力
    * t: 第一个ghost模块的通道数大小
    * c: 瓶颈结构最终输出通道数的大小
    * SE: 是否使用注意力模块
    * s: 步长大小,如果为2,则会进行高和宽的压缩
    """
    cfgs = [
        # k,  t,   c,  SE,  s
        # stage1
        [[3,  16,  16,  0,  1]],      # 112 x 112 x 16 --> 112 x 112 x 16
        # stage2
        [[3,  48,  24,  0,  2]],      # 112 x 112 x 16 --> 56 x 56 x 24
        [[3,  72,  24,  0,  1]],
        # stage3
        [[5,  72,  40, 0.25, 2]],     # 56 x 56 x 24 --> 28 x 28 x 40
        [[5, 120,  40, 0.25, 1]],
        # stage4
        [[3, 240,  80,  0,  2]],      # 28 x 28 x 40 --> 14 x 14 x 80
        [[3, 200,  80,  0,  1],
         [3, 184,  80,  0,  1],
         [3, 184,  80,  0,  1],
         [3, 480, 112, 0.25, 1],      # 14 x 14 x 80 --> 14 x 14 x 112
         [3, 672, 112, 0.25, 1]],
        # stage5
        [[5, 672, 160, 0.25, 2]],     # 14 x 14 x 112 --> 7 x 7 x 160
        [[5, 960, 160,  0,  1],
         [5, 960, 160, 0.25, 1],
         [5, 960, 160,  0,  1],
         [5, 960, 160, 0.25, 1]]
    ]
    return GhostNet(cfgs, num_classes=num_classes, block=GhostBottleneckv1)

def ghostnetv2(num_classes):
    """
    * k: 卷积核大小,表示跨特征点提取特征的能力
    * t: 第一个ghost模块的通道数大小
    * c: 瓶颈结构最终输出通道数的大小
    * SE: 是否使用注意力模块
    * s: 步长大小,如果为2,则会进行高和宽的压缩
    """
    cfgs = [
        # k,   t,  c,  SE,  s
        # stage1
        [[3,  16,  16,  0,  1]],      # 112 x 112 x 16 --> 112 x 112 x 16
        # stage2
        [[3,  48,  24,  0,  2]],      # 112 x 112 x 16 --> 56 x 56 x 24
        [[3,  72,  24,  0,  1]],
        # stage3
        [[5,  72,  40, 0.25, 2]],     # 56 x 56 x 24 --> 28 x 28 x 40
        [[5, 120,  40, 0.25, 1]],
        # stage4
        [[3, 240,  80,  0,  2]],      # 28 x 28 x 40 --> 14 x 14 x 80
        [[3, 200,  80,  0,  1],
         [3, 184,  80,  0,  1],
         [3, 184,  80,  0,  1],
         [3, 480, 112, 0.25, 1],      # 14 x 14 x 80 --> 14 x 14 x 112
         [3, 672, 112, 0.25, 1]],
        # stage5
        [[5, 672, 160, 0.25, 2]],     # 14 x 14 x 112 --> 7 x 7 x 160
        [[5, 960, 160,  0,  1],
         [5, 960, 160, 0.25, 1],
         [5, 960, 160,  0,  1],
         [5, 960, 160, 0.25, 1]]
    ]
    return GhostNet(cfgs, num_classes=num_classes, block=GhostBottleneckv2)



if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = ghostnetv2(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # v1 Total params: 3,908,608
    # v2 Total params: 4,883,008


import math
from decimal import Decimal

__all__ = ["sqrt", "rsqrt", "odd_factorial", "even_factorial", "factorial", "pow", "retain",
           "round_down", "round_up", "Sum", "Subtract", "Multiply", "Divide",
           "EuclidDistance", "ChessboardDistance", "CityblockDistance", "CenterPoint"]

def sqrt(x):
    """二次方根"""
    return pow(x, 2)

def rsqrt(x):
    """二次方根倒数"""
    return 1 / sqrt(x)

def odd_factorial(n):
    # 奇数的阶乘
    result = 1
    for i in range(1, 2 * n + 1, 2):
        result *= i
    return result

def even_factorial(n):
    # 偶数的阶乘
    result = 1
    for i in range(2, 2 * n + 1, 2):
        result *= i
    return result

def factorial(n):
    # 阶乘
    return math.factorial(n)

def pow(x, n):
    """幂函数计算"""
    return x ** n

def retain(val, t=2):
    """精准保留小数,默认2位"""
    value = Decimal(val)
    return round(value, t)

def round_up(x):
    """向上取整"""
    return math.ceil(x)

def round_down(x):
    """向下取整"""
    return math.floor(x)

def Sum(*args):
    """计算参数的和,支持多个参数或一个一维列表(元组)"""
    if not args:
        raise ValueError("At least one argument is required.")

    if len(args) == 1 and (isinstance(args[0], list) or isinstance(args[0], tuple)):
        return sum(args[0])
    else:
        return sum(args)

def Subtract(*args):
    """计算参数的差,支持多个参数或一个一维列表(元组)"""
    if not args:
        raise ValueError("At least one argument is required.")

    if len(args) == 1 and (isinstance(args[0], list) or isinstance(args[0], tuple)):
        result = args[0][0] - sum(args[0][1:])
    else:
        result = args[0] - sum(args[1:])
    return result

def Multiply(*args):
    """计算参数的乘积,支持多个参数或一个一维列表(元组)"""
    if not args:
        raise ValueError("At least one argument is required.")

    if len(args) == 1 and (isinstance(args[0], list) or isinstance(args[0], tuple)):
        result = 1
        for arg in args[0]:
            result *= arg
    else:
        result = 1
        for arg in args:
            result *= arg
    return result

def Divide(*args):
    """计算参数的商,支持多个参数或一个一维列表(元组)"""
    if len(args) == 1 and (isinstance(args[0], list) or isinstance(args[0], tuple)):
        result = args[0][0]
        for arg in args[0][1:]:
            if arg == 0:
                raise ValueError("Cannot divide by zero.")
            result /= arg
    else:
        result = args[0]
        for arg in args[1:]:
            if arg == 0:
                raise ValueError("Cannot divide by zero.")
            result /= arg
    return result

def EuclidDistance(point1, point2):
    """欧式距离, 即两点间距离公式"""
    return math.sqrt((point2[0] - point1[0]) ** 2 + (point2[1] - point1[1]) ** 2)

def ChessboardDistance(point1, point2):
    """棋盘距离"""
    return abs(point2[0] - point1[0]) + abs(point2[1] - point1[1])

def CityblockDistance(point1, point2):
    """Cityblock距离"""
    return max(abs(point2[0] - point1[0]), abs(point2[1] - point1[1]))

def CenterPoint(point1, point2):
    """中心点计算公式"""
    return ((point1[0] + point2[0]) / 2, (point1[1] + point2[1]) / 2)


"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is used to quantify two-dimensional information such as the number, length, width, and area of cracks
in the semantic segmentation graph of neural networks.
"""
import cv2
import numpy as np

import string
import os
from skimage.filters import threshold_otsu,median
from skimage.morphology import skeletonize,dilation,disk
from skimage import io, morphology, measure

from pyzjr.augmentation.Pixel import SkeletonMap, incircle
from pyzjr.augmentation.Color import putBoxText


# 裂缝类型
class CrackType():
    """直方图投影法推断裂缝类型,仅仅能区分线性裂缝和网状裂缝"""
    def __init__(self, threshold=3, HWratio=10, Histratio=0.5):
        """
        初始化分类裂缝的参数
        :param threshold: 阈值，用于分类裂缝的阈值
        :param HWratio: 高宽比，用于分类裂缝的高宽比阈值
        :param Histratio: 直方图比例，用于分类裂缝的直方图比例阈值
        """
        self.threshold = threshold
        self.HWratio = HWratio
        self.Histratio = Histratio
        self.types = {0: 'Horizontal',
                      1: 'Vertical',
                      2: 'Oblique',
                      3: 'Mesh'}

    def inference_minAreaRect(self, minAreaRect):
        """
        旋转矩形框长边与x轴的夹角.
        旋转角度 angle 是相对于图像水平方向的夹角，范围是 -90 到 +90 度.
        然而，一般情况下，我们习惯将角度定义为相对于 x 轴正方向的夹角，范围是 -180 到 +180 度.
        """
        w, h = minAreaRect[1]
        if w > h:
            angle = int(minAreaRect[2])
        else:
            angle = -(90 - int(minAreaRect[2]))
        return w, h, angle

    def classify(self, minAreaRect, skeleton_pts, HW):
        """
        针对当前crack instance，对其进行分类；
        主要利用了骨骼点双向投影直方图、旋转矩形框宽高比/角度；
        :param minAreaRect: 最小外接矩形框，[(cx, cy), (w, h), angle]；
        :param skeleton_pts: 骨骼点坐标；
        :param HW: 当前patch的高、宽；
        """
        H, W = HW
        w, h, angle = self.inference_minAreaRect(minAreaRect)
        if w / h < self.HWratio or h / w < self.HWratio:
            pts_y, pts_x = skeleton_pts[:, 0], skeleton_pts[:, 1]
            hist_x = np.histogram(pts_x, W)
            hist_y = np.histogram(pts_y, H)
            if self.hist_judge(hist_x[0]) and self.hist_judge(hist_y[0]):
                return 3

        return self.angle2cls(angle)

    def hist_judge(self, hist_v):
        less_than_T = np.count_nonzero((hist_v > 0) & (hist_v <= self.threshold))
        more_than_T = np.count_nonzero(hist_v > self.threshold)
        return more_than_T / (less_than_T + 1e-5) > self.Histratio

    @staticmethod
    def angle2cls(angle):
        angle = abs(angle)
        assert 0 <= angle <= 90, "ERROR: The angle value exceeds the limit and should be between 0 and 90 degrees!"
        if angle < 35:
            return 0
        elif 35 <= angle <= 55:
            return 2
        elif angle > 55:
            return 1
        else:
            return None

def _get_minAreaRect_information(mask):
    """
    从二值化掩膜图像中获取最小外接矩形的相关信息
    :param mask:二值化掩膜图像，包含目标区域的白色区域
    :return:最小外接矩形的信息，包括中心点坐标、宽高和旋转角度
    """
    gray_image = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)
    _, mask = cv2.threshold(gray_image, 127, 255, cv2.THRESH_BINARY)
    contours, hierarchy = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    contour_merge = np.vstack(contours)
    minAreaRect = cv2.minAreaRect(contour_merge)
    return minAreaRect

def infertype(mask):
    """推导裂缝类型"""
    crack = CrackType()
    H, W = mask.shape[:2]
    minAreaRect = _get_minAreaRect_information(mask)
    skeimage, skepoints = SkeletonMap(mask)
    result = crack.classify(minAreaRect, skepoints, HW=(H, W))
    crack_type = crack.types[result]
    return crack_type

class Crack_areas_and_numbers():
    def __init__(self, mask, t=2):
        self.imgMask = mask.copy()
        gray_image = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)
        _, self.thresh = cv2.threshold(gray_image, 127, 255, cv2.THRESH_BINARY)
        self.Bboxing=[]
        self.t = t

    def Totalareas(self):
        """目标像素白色的个数"""
        image_array = np.array(self.thresh)
        white_pixels = np.count_nonzero(image_array)
        return white_pixels

    def Eachareas(self, merge_threshold=3, area_threshold=50):
        """每条裂缝的面积,并用大写字母来进行标记"""
        connected_image = morphology.closing(self.thresh, morphology.disk(merge_threshold))
        labeled_image = measure.label(connected_image, connectivity=2)
        region_props = measure.regionprops(labeled_image)
        self.area = {}
        crack_label = ord('A')
        for region in region_props:
            area_value = region.area
            if area_value >= area_threshold:
                minr, minc, maxr, maxc = region.bbox
                self.Bboxing.append([(minc, minr), (maxc, maxr)])
                putBoxText(self.imgMask, [minc, minr, maxc, maxr],chr(crack_label))
                if crack_label <= ord('Z'):
                    self.area[chr(crack_label)] = area_value
                    crack_label += 1
        return self.area, self.Bboxing, self.imgMask

    def Cracknums(self):
        """裂缝个数"""
        return len(self.area)

def Pixel_Coords(img, val=255):
    """用于计算所有目标像素的坐标位置,默认为白色"""
    gray_image = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    _, thresh = cv2.threshold(gray_image, 127, 255, cv2.THRESH_BINARY)
    y_coords, x_coords = np.where(thresh == val)
    coords_list = list(zip(x_coords, y_coords))
    return coords_list

def reorder(coord, type):
    """np.where默认是按照从上到下进行查找的"""
    if type == "Horizontal":
        coords = sorted(coord, key=lambda x: x[0])
        return coords
    elif type == "Vertical" or type == "Oblique":
        return coord

def Crop_cracks_img(mask, Bboxing):
    """根据边界框,进行裁剪,并存入列表中"""
    cropped_cracks = []
    for bbox in Bboxing:
        x1, y1 = bbox[0]
        x2, y2 = bbox[1]
        cropped_crack = mask[y1:y2, x1:x2]
        cropped_cracks.append(cropped_crack)
    return cropped_cracks

def Crack_label(cracknums):
    """创建一个关于A-Z的标签"""
    labels = []
    for i in range(cracknums):
        label = string.ascii_uppercase[i]
        labels.append(label)
    return labels

def Nan():
    """仅表示一个无效值"""
    return 'Nan'

def Classify_points_basedon_Bbox(gujia_pos, Bboxing):
    """按照边界框对像素进行分类"""
    classified_gujia_pos = {}
    for point in gujia_pos:
        x, y = point
        for i, bbox in enumerate(Bboxing):
            (minc, minr), (maxc, maxr) = bbox
            if minc <= x <= maxc and minr <= y <= maxr:
                category_label = chr(ord('A') + i)
                if category_label not in classified_gujia_pos:
                    classified_gujia_pos[category_label] = []
                classified_gujia_pos[category_label].append(point)
                break
    return classified_gujia_pos

def Crack_of_length(coords_lst, bias = 1.118):
    """这里用于计算裂缝的长度"""
    connection, inflexion, spaceBetween = 0, 0, 0
    for i in range(len(coords_lst) - 1):
        start = coords_lst[i]
        end = coords_lst[i + 1]
        gap = ((end[0] - start[0]) ** 2 + (end[1] - start[1]) ** 2) ** 0.5
        if gap == 1.0:
            connection += 1
        elif gap == np.sqrt(2):
            inflexion += 1
        spaceBetween += gap
    return spaceBetween + bias, inflexion, connection

def Crack_of_width(total_area, total_length):
    """求解裂缝平均宽度"""
    ave_width = total_area / (total_length + 1e-5)
    return ave_width

class Crack_skeleton():
    def sketioncv(self, single_pic):
        image = cv2.imread(single_pic, cv2.IMREAD_GRAYSCALE)
        _, binary = cv2.threshold(image, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
        binary = cv2.dilate(binary, kernel, iterations=1)
        binary = cv2.medianBlur(binary, 5)
        binary = cv2.dilate(binary, kernel, iterations=1)
        binary = cv2.medianBlur(binary, 5)
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
        binary = cv2.morphologyEx(binary, cv2.MORPH_CLOSE, kernel)
        binary=cv2.cvtColor(binary,cv2.COLOR_BGR2RGB)
        skeleton = skeletonize(binary)
        skeleton=cv2.cvtColor(skeleton,cv2.COLOR_RGB2BGR)
        return skeleton

    def sketionio(self,single_pic):
        image = io.imread(single_pic, as_gray=True)
        thresh = threshold_otsu(image)
        binary = image > thresh

        binary = dilation(binary, disk(3))
        binary = median(binary, footprint=morphology.disk(5))
        binary = dilation(binary, disk(2))
        binary = median(binary, footprint=morphology.disk(5))

        selem = morphology.disk(3)
        binary = morphology.closing(binary, selem)

        skeleton = skeletonize(binary)
        return skeleton

    def sketions(self, input_folder='num', output_folder='output'):
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)  # 如果输出文件夹不存在，就创建它
        for filename in os.listdir(input_folder):
            if filename.endswith('.jpg') or filename.endswith('.png'):
                input_file=os.path.join(input_folder, filename)
                output_filename = os.path.join(output_folder, filename)
                skeleton=self.sketionio(input_file)
                io.imsave(output_filename, skeleton)
        return output_folder

def DetectCracks(mask, skeimg, mode, merge_threshold=3, area_threshold=50, showcrackinformation=True):
    crack_area = Crack_areas_and_numbers(mask)
    cracktype = infertype(mask)
    if cracktype == "Mesh":
        total_areas = crack_area.Totalareas()
        total_length = width = CrackNum = crack_information = Nan()
        return CrackNum, total_areas, total_length, width, crack_information
    else:
        area, Bboxing, Maskimg = crack_area.Eachareas(merge_threshold=merge_threshold,area_threshold=area_threshold)
        total_area = sum(area.values())
        CrackNum = crack_area.Cracknums()
        label = Crack_label(CrackNum)
        cropped_cracks = Crop_cracks_img(mask, Bboxing)

        gujia_pos = Pixel_Coords(skeimg)
        classified_gujia_pos = Classify_points_basedon_Bbox(gujia_pos, Bboxing)
        crack_information=[]
        all_length=[]
        for i, crop in enumerate(cropped_cracks):
            crop_type = infertype(crop)
            new_classified_gujia_pos = reorder(classified_gujia_pos[label[i]], crop_type)
            spaceBetween, inflexion, connection = Crack_of_length(new_classified_gujia_pos)
            all_length.append(spaceBetween)
            crack_information.append([f"裂缝{label[i]},面积{area[label[i]]},长度{spaceBetween},拐点:{inflexion},连通:{connection}"])
            if showcrackinformation:
                print(f"{label[i]},裂缝面积为{area[label[i]]},裂缝长度:{spaceBetween},拐点:{inflexion},连通:{connection}")
            i += 1
        total_length = sum(all_length)
        if mode == 0:
            # 平均宽度
            width = Crack_of_width(total_area, total_length)
            return CrackNum, total_area, total_length, width, crack_information
        elif mode == 1:
            # 最大内切圆
            gray_image = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)
            _, thresh = cv2.threshold(gray_image, 127, 255, cv2.THRESH_BINARY)
            contours_arr, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
            mask_gray = cv2.cvtColor(mask,cv2.COLOR_BGR2GRAY)
            result, labelwidth = incircle(mask_gray, contours_arr)
            return CrackNum, total_area, total_length, labelwidth, crack_information

if __name__=="__main__":
    path = r"D:\pythonprojects\RoadCrack\dimension2_data\num\021.png"

    mask = cv2.imread(path)
    ske = Crack_skeleton()
    skeimg = ske.sketioncv(path)
    CrackNum, total_area, total_length, labelwidth, crack_information = DetectCracks(mask, skeimg, mode=1)

    print(f"裂缝个数为：{CrackNum}\n"
          f"裂缝总面积为：{total_area}\n"
          f"裂缝总长度为：{total_length}\n"
          f"裂缝宽度为：{labelwidth}\n"
          f"裂缝细致信息为：{crack_information}")


import cv2
import time
import imageio
import numpy as np

__all__ = [
            "VideoCap",
            "FPS",
            "Timer",
            "Mp4toGif",
           ]

class VideoCap():
    """
    Customized Python video reading class
    """
    def CapInit(self, mode=0, w=640, h=480, l=150):
        self.cap = cv2.VideoCapture(mode)
        self.cap.set(3, w)
        self.cap.set(4, h)
        self.cap.set(10, l)

    def read(self, flip=None):
        """
        :param flip: -1: Horizontal and vertical directions,
                      0: along the y-axis, vertical,
                      1: along the x-axis, horizontal
        """
        _, img = self.cap.read()
        if flip is not None:
            if flip in [-1, 0, 1]:
                img = cv2.flip(img, flip)
            else:
                raise ValueError("[VideoCap]:The flip parameter should be -1, 0, or 1")
        return img

    def free(self):
        """
        Release camera
        """
        self.cap.release()

class FPS:
    def __init__(self):
        self.pTime = time.time()

    def update(self, img=None, pos=(20, 50), color=(255, 0, 0), scale=3, thickness=3):
        """
        更新帧速率
        :param img: The displayed image can be left blank if only the fps value is needed
        :param pos: Position on FPS on image
        :param color: The color of the displayed FPS value
        :param scale: The proportion of displayed FPS values
        :param thickness: The thickness of the displayed FPS value
        :return:
        """
        cTime = time.time()
        try:
            fps = 1 / (cTime - self.pTime)
            self.pTime = cTime
            if img is None:
                return fps
            else:
                cv2.putText(img, f'FPS: {int(fps)}', pos, cv2.FONT_HERSHEY_PLAIN,
                            scale, color, thickness)
                return fps, img
        except:
            return 0

class Timer:
    def __init__(self):
        """Start is called upon creation"""
        self.times = []
        self.start()
    def start(self):
        """initial time"""
        self.gap = time.time()
    def stop(self):
        """The time interval from the start of timing to calling the stop method"""
        self.times.append(time.time() - self.gap)
        return self.times[-1]
    def avg(self):
        """Average time consumption"""
        return sum(self.times) / len(self.times)
    def total(self):
        """Total time consumption"""
        return sum(self.times)
    def cumsum(self):
        """Accumulated sum of time from the previous n runs"""
        return np.array(self.times).cumsum().tolist()


def Mp4toGif(mp4, name='result.gif', fps=10, start=None, end=None):

    cap = cv2.VideoCapture(mp4)
    all_images = []
    frame_count = 0

    while True:
        ret, img = cap.read()
        if ret is False:
            break
        if start is not None and frame_count < start:
            frame_count += 1
            continue
        if end is not None and frame_count >= end:
            break
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        all_images.append(img)
        frame_count += 1

    duration = int(1000 / fps)  # 将帧率转换为每帧之间的延迟时间（毫秒）
    gif = imageio.mimsave(name, all_images, duration=duration)
    print("转换完成！")

if __name__ == "__main__":
    import pyzjr.Z as Z
    fpsReader = FPS()
    Vcap = VideoCap()
    Vcap.CapInit(mode=0)
    while True:
        img = Vcap.read()
        fps, img = fpsReader.update(img)
        cv2.imshow("Image", img)
        k = cv2.waitKey(1)
        if k == Z.Esc:
            break


"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is used to implement clarity evaluation algorithms.
"""
import cv2
import math
import numpy as np
import scipy.ndimage as ndi
from skimage.filters import sobel

import pyzjr.Z as Z
from pyzjr.data.reader import getPhotopath

__all__ = ["ImgDefinition", "Fuzzy_image", "vagueJudge",]

class ImgDefinition():
    """
    清晰度评价函数: https://blog.csdn.net/m0_62919535/article/details/127818006
    Laplacian与Gradientmetric最靠谱
    """
    def Brenner(self, img):
        '''
        Brenner梯度函数
        :param img:narray 二维灰度图像
        :return: int 图像越清晰越大
        '''
        shapes = np.shape(img)
        output = 0
        for x in range(0, shapes[0] - 2):
            for y in range(0, shapes[1]):
                output += (int(img[x + 2, y]) - int(img[x, y])) ** 2
        return output

    def EOG(self, img):
        '''
        能量梯度函数(Energy of Gradient)
        :param img:narray 二维灰度图像
        :return: int 图像越清晰越大
        '''
        shapes = np.shape(img)
        output = 0
        for x in range(0, shapes[0] - 1):
            for y in range(0, shapes[1] - 1):
                output += ((int(img[x + 1, y]) - int(img[x, y])) ** 2 + (int(img[x, y + 1]) - int(img[x, y])) ** 2)
        return output

    def Roberts(self,img):
        '''
        Roberts函数
        :param img:narray 二维灰度图像
        :return: int 图像越清晰越大
        '''
        shapes = np.shape(img)
        output = 0
        for x in range(0, shapes[0] - 1):
            for y in range(0, shapes[1] - 1):
                output += ((int(img[x + 1, y + 1]) - int(img[x, y])) ** 2 + (
                        int(img[x + 1, y]) - int(img[x, y + 1])) ** 2)
        return output

    def Laplacian(self,img):
        '''
        Laplace
        :param img:narray 二维灰度图像
        :return: int 图像越清晰越大
        '''
        return cv2.Laplacian(img, Z.Lap_64F).var()

    def SMD(self,img):
        '''
        SMD(灰度方差)函数
        :param img:narray 二维灰度图像
        :return: int 图像越清晰越大
        '''
        shape = np.shape(img)
        output = 0
        for x in range(1, shape[0] - 1):
            for y in range(0, shape[1]):
                output += math.fabs(int(img[x, y]) - int(img[x, y - 1]))
                output += math.fabs(int(img[x, y] - int(img[x + 1, y])))
        return output

    def SMD2(self,img):
        '''
        （灰度方差乘积）函数
        :param img:narray 二维灰度图像
        :return: int 图像约清晰越大
        '''
        shape = np.shape(img)
        output = 0
        for x in range(0, shape[0] - 1):
            for y in range(0, shape[1] - 1):
                output += math.fabs(int(img[x, y]) - int(img[x + 1, y])) * math.fabs(
                    int(img[x, y] - int(img[x, y + 1])))
        return output

    def Gradientmetric(self,img, size=11):
        """
        计算指示图像中模糊强度的度量(0表示无模糊, 1表示最大模糊)
        但实际上还是要根据我们的标准图与模糊图得到模糊区间
        :param img: 单通道进行处理,灰度图
        :param size: 重新模糊过滤器的大小
        :return: float,唯一不是值越大,越清晰的算法
        """
        image = np.array(img, dtype=np.float32) / 255
        n_axes = image.ndim
        shape = image.shape
        blur_metric = []

        slices = tuple([slice(2, s - 1) for s in shape])
        for ax in range(n_axes):
            filt_im = ndi.uniform_filter1d(image, size, axis=ax)
            im_sharp = np.abs(sobel(image, axis=ax))
            im_blur = np.abs(sobel(filt_im, axis=ax))
            T = np.maximum(0, im_sharp - im_blur)
            M1 = np.sum(im_sharp[slices])
            M2 = np.sum(T[slices])
            blur_metric.append(np.abs((M1 - M2)) / M1)

        return np.max(blur_metric) if len(blur_metric) > 0 else 0.0

    def get_method(self, mode):
        methods = {
            "B": self.Brenner,
            "E": self.EOG,
            "R": self.Roberts,
            "L": self.Laplacian,
            "S": self.SMD,
            "S2": self.SMD2,
            "G": self.Gradientmetric,
        }
        return methods.get(mode, None)

def definemode(img, mode):
    ti = ImgDefinition()
    method = ti.get_method(mode)

    if method is not None:
        return method(img)
    else:
        raise ValueError(f"[pyzjr]:Invalid mode: {mode}")

class Fuzzy_image():
    def __init__(self, mode="L"):
        """
        模糊判定区间: https://blog.csdn.net/m0_62919535/article/details/128061017
        :param mode: 图像清晰度算法模式,建议采用“L”或“G”算法
        """
        self.mode=mode
    def getImgVar(self, image):
        """
        :param image: 图像
        :return: 图片清晰度数值
        """
        imggray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        imageVar = definemode(imggray,mode=self.mode)
        return abs(imageVar)

    def getReorder(self, imgfile, reverse = False):
        """
        :param imgfile: getPhotopath()获取图片路径的列表
        :param reverse: 表示是否进行反转排序(从大到小)
        :return: 模糊值排序
        """
        c = []
        for i in imgfile:
            img = cv2.imread(i)
            image = self.getImgVar(img)
            c.append(float(f"{image:.3f}"))
        c.sort(reverse=reverse)
        return c

    def getInterval(self, pathVague, pathStd):
        """
        :param pathVague:模糊图的数据集文件夹位置
        :param pathStd: 标准图的数据文件夹位置
        :return: 模糊判定区间
        """
        imgfile1, _ = getPhotopath(pathVague,debug=False)
        imgfile2, _ = getPhotopath(pathStd,debug=False)
        a = self.getReorder(imgfile1)
        b = self.getReorder(imgfile2)
        if self.mode=="G":
            thr = (b[-1], a[0])
        else:
            thr = (a[-1], b[0])
        return thr

def vagueJudge(img, pathVague, pathStd, mode="L", show_minandmax=True):
    """
    :param img: 图片
    :param show_minandmax:是否输出模糊未知区间
    :return: 模糊数值打印在图片上显示,只要没有超过清晰图像的阈值，全部判断为模糊
    """
    Fuzzy = Fuzzy_image(mode)
    imgVar = Fuzzy.getImgVar(img)
    minThr, maxThr = Fuzzy.getInterval(pathVague=pathVague, pathStd=pathStd)
    img2 = img.copy()
    if imgVar > maxThr:
        if mode == "G":
            text = "Vague"
            color = Z.red
        else:
            text = "Not Vague"
            color = Z.blue
    else:
        if mode == "G":
            text = "Not Vague"
            color = Z.blue
        else:
            text = "Vague"
            color = Z.red

    cv2.putText(img2, f"{text}{imgVar:.2f}", (12, 70), cv2.FONT_HERSHEY_PLAIN, 3, color, 3)

    if show_minandmax:
        print(minThr, maxThr)
    else:
        pass
    return img2

if __name__=="__main__":
    path = r"ces\Standards\001.jpg"   # 需要进行测试的图片
    path2 = r"ces\test\02.jpg"
    pathVague = r"ces\test"           # 模糊图像数据
    pathStd = r"ces\Standards"        # 标准图像数据
    img = cv2.imread(path)
    test = vagueJudge(img, pathVague, pathStd, mode="G")
    cv2.imshow("test", test)
    cv2.waitKey(0)


from pyzjr.videos.utils import Timer
from pyzjr.data.printf import colorstr
import time
import torch
import argparse
import torch.nn as nn
try:
    import thop  # for FLOPs computation
except ImportError:
    thop = None

__all__ = ["Runcodes", "LoadingBar", "show_config", "GPU_INFO", "time_sync", "profile"]

class Runcodes:
    """
    Comparing the running time of different algorithms.
    example:
        with Runcodes("inference time"):
            output = ...
    """
    def __init__(self, description='Done'):
        self.description = description

    def __enter__(self):
        self.timer = Timer()
        return self

    def __exit__(self, *args):
        print(f'{self.description}: {self.timer.stop():.5f} sec')


class LoadingBar:
    def __init__(self, length: int = 40):
        """Simple dynamic display bar
        example:
            loading_bar = LoadingBar()
            for i in range(101):
                progress = i / 100.0
                bar_string = loading_bar(progress)
                print(f"\r[{bar_string}] {i}% ", end="")
                time.sleep(0.1)  # 模拟加载延迟
            print("\nLoading Complete!")
        """
        self.length = length
        self.symbols = ['┈', '░', '▒', '▓']

    def __call__(self, progress: float) -> str:
        p = int(progress * self.length*4 + 0.5)
        d, r = p // 4, p % 4
        return '┠┈' + d * '█' + ((self.symbols[r]) + max(0, self.length-1-d) * '┈' if p < self.length*4 else '') + "┈┨"

def show_config(head="Configurations", args=None, color='black', **kwargs):
    """显示配置信息"""
    print(colorstr(color, f'{head}:'))
    print(colorstr(color, '-' * 113))
    print(colorstr(color, '|%5s | %45s | %55s|' % ('order', 'keys', 'values')))
    print(colorstr(color, '-' * 113))
    counter = 0
    if args is not None:
        if isinstance(args, argparse.Namespace):
            config_dict = vars(args)
            for key, value in config_dict.items():
                counter += 1
                print(colorstr(color, f'|%5d | %45s | %55s|' % (counter, key, value)))
        elif isinstance(args, list):
            for arg in args:
                counter += 1
                print(colorstr(color, f'|%5d | %45s | {"": >55}|' % (counter, arg)))  # Assuming each element in the list is a key
        else:
            counter += 1
            print(colorstr(color, f'|%5d | %45s | %55s|' % (counter, "arg", args)))  # Assuming args is a single value

    for key, value in kwargs.items():
        counter += 1
        print(colorstr(color, f'|%5d | %45s | %55s|' % (counter, key, value)))

    print(colorstr(color, '-' * 113))

def GPU_INFO(headColor="red",gpuColor="red"):
    print(colorstr(headColor, 'GPU INFO:'))
    print(colorstr(gpuColor, torch.cuda.memory_summary()), end='')

def profile(input, ops, n=10, cuda=True):
    """
    Usage:
        model = MyModel().to('cuda')
        input_tensor = torch.randn(1, 3, 512, 512).to('cuda')
        profile(input_tensor, model, n=10)
    """
    results = []
    print(f"{'Params':>12s}{'GFLOPs':>12s}{'GPU_mem (GB)':>14s}{'forward (ms)':>14s}{'backward (ms)':>14s}"
          f"{'input':>24s}{'output':>24s}")
    if torch.cuda.is_available():
        device = "cuda" if cuda else "cpu"
    else:
        device = "cpu"
    for x in input if isinstance(input, list) else [input]:
        x = x.to(device)
        x.requires_grad = True
        for m in ops if isinstance(ops, list) else [ops]:
            m = m.to(device) if hasattr(m, 'to') else m
            m = m.half() if hasattr(m, 'half') and isinstance(x, torch.Tensor) and x.dtype is torch.float16 else m
            tf, tb, t = 0, 0, [0, 0, 0]  # dt forward, backward
            try:
                flops = thop.profile(m, inputs=(x, ), verbose=False)[0] / 1E9 * 2  # GFLOPs
            except Exception:
                flops = 0
            try:
                for _ in range(n):
                    t[0] = time_sync()
                    y = m(x)
                    t[1] = time_sync()
                    try:
                        _ = (sum(yi.sum() for yi in y) if isinstance(y, list) else y).sum().backward()
                        t[2] = time_sync()
                    except Exception as e:  # no backward method
                        print(e)
                        t[2] = float('nan')
                    tf += (t[1] - t[0]) * 1000 / n  # ms per op forward
                    tb += (t[2] - t[1]) * 1000 / n  # ms per op backward
                mem = torch.cuda.memory_reserved() / 1E9 if torch.cuda.is_available() else 0  # (GB)
                s_in, s_out = (tuple(x.shape) if isinstance(x, torch.Tensor) else 'list' for x in (x, y))  # shapes
                p = sum(x.numel() for x in m.parameters()) if isinstance(m, nn.Module) else 0  # parameters
                print(f'{p:12}{flops:12.4g}{mem:>14.3f}{tf:14.4g}{tb:14.4g}{str(s_in):>24s}{str(s_out):>24s}')
                results.append([p, flops, mem, tf, tb, s_in, s_out])
            except Exception as e:
                print(e)
                results.append(None)
            torch.cuda.empty_cache()
    return results

def time_sync():
    # PyTorch-accurate time
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    return time.time()

if __name__=="__main__":
    import platform
    import sys

    show_config(Python_vision=".".join(map(str, [sys.version_info.major, sys.version_info.minor, sys.version_info.micro])),
                CUDA=torch.version.cuda,
                CUDA_device=torch.cuda.get_device_name(0) if torch.cuda.is_available() else None,
                Torch_vision=torch.__version__,
                os_info=platform.system() + " " + platform.version(),
                color='red',
                )


"""
Copyright (c) 2024, Auorui.
All rights reserved.

The Torch implementation of average pooling and maximum pooling has been compared with the official Torch implementation
Time: 2024-01-22  17:28
"""
import torch
import torch.nn as nn

class MaxPool2d(nn.Module):
    """
    池化层计算公式:
        output_size = [(input_size−kernel_size) // stride + 1]
    """
    def __init__(self, kernel_size, stride):
        super(MaxPool2d, self).__init__()
        self.kernel_size = kernel_size
        self.stride = stride

    def max_pool2d(self, input_tensor, kernel_size, stride):
        batch_size, channels, height, width = input_tensor.size()
        output_height = (height - kernel_size) // stride + 1
        output_width = (width - kernel_size) // stride + 1
        output_tensor = torch.zeros(batch_size, channels, output_height, output_width)

        for i in range(output_height):
            for j in range(output_width):
                # 获取输入张量中与池化窗口对应的部分
                window = input_tensor[:, :,
                         i * stride: i * stride + kernel_size, j * stride: j * stride + kernel_size]
                output_tensor[:, :, i, j] = torch.max(window.reshape(batch_size, channels, -1), dim=2)[0]
        return output_tensor

    def forward(self, input_tensor):
        return self.max_pool2d(input_tensor, kernel_size=self.kernel_size, stride=self.stride)


class AvgPool2d(nn.Module):
    """
    池化层计算公式:
        output_size = [(input_size−kernel_size) // stride + 1]
    """
    def __init__(self, kernel_size, stride):
        super(AvgPool2d, self).__init__()
        self.kernel_size = kernel_size
        self.stride = stride

    def avg_pool2d(self, input_tensor, kernel_size, stride):
        batch_size, channels, height, width = input_tensor.size()
        output_height = (height - kernel_size) // stride + 1
        output_width = (width - kernel_size) // stride + 1
        output_tensor = torch.zeros(batch_size, channels, output_height, output_width)

        for i in range(output_height):
            for j in range(output_width):
                # 获取输入张量中与池化窗口对应的部分
                window = input_tensor[:, :,
                         i * stride: i * stride + kernel_size, j * stride:j * stride + kernel_size]
                output_tensor[:, :, i, j] = torch.mean(window.reshape(batch_size, channels, -1), dim=2)
        return output_tensor

    def forward(self, input_tensor):
        return self.avg_pool2d(input_tensor, kernel_size=self.kernel_size, stride=self.stride)


if __name__=="__main__":
    # input_data = torch.rand((1, 3, 3, 3))
    input_data = torch.Tensor([[[[0.3939, 0.8964, 0.3681],
                               [0.5134, 0.3780, 0.0047],
                               [0.0681, 0.0989, 0.5962]],
                              [[0.7954, 0.4811, 0.3329],
                               [0.8804, 0.3986, 0.3561],
                               [0.2797, 0.3672, 0.6508]],
                              [[0.6309, 0.1340, 0.0564],
                               [0.3101, 0.9927, 0.5554],
                               [0.0947, 0.2305, 0.8299]]]])

    print(input_data.shape)

    kernel_size = 3
    stride = 1
    MaxPool2d1 = nn.MaxPool2d(kernel_size, stride)
    output_data_with_torch_max = MaxPool2d1(input_data)
    AvgPool2d1 = nn.AvgPool2d(kernel_size, stride)
    output_data_with_torch_avg = AvgPool2d1(input_data)
    AvgPool2d2 = AvgPool2d(kernel_size, stride)
    output_data_with_torch_Avg = AvgPool2d2(input_data)
    MaxPool2d2 = MaxPool2d(kernel_size, stride)
    output_data_with_torch_Max = MaxPool2d2(input_data)
    # output_data_with_max = max_pool2d(input_data, kernel_size, stride)
    # output_data_with_avg = avg_pool2d(input_data, kernel_size, stride)

    print("\ntorch.nn pooling Output:")
    print(output_data_with_torch_max,"\n",output_data_with_torch_max.size())
    print(output_data_with_torch_avg,"\n",output_data_with_torch_avg.size())
    print("\npooling Output:")
    print(output_data_with_torch_Max,"\n",output_data_with_torch_Max.size())
    print(output_data_with_torch_Avg,"\n",output_data_with_torch_Avg.size())
    # 直接使用bool方法判断会因为浮点数的原因出现偏差
    print(torch.allclose(output_data_with_torch_max,output_data_with_torch_Max))
    print(torch.allclose(output_data_with_torch_avg,output_data_with_torch_Avg))
    # tensor([[[[0.8964]],       # output_data_with_max
    #          [[0.8804]],
    #          [[0.9927]]]])
    # tensor([[[[0.3686]],       # output_data_with_avg
    #           [[0.5047]],
    #           [[0.4261]]]])


"""
Copyright (c) 2024, Auorui.
All rights reserved.
Time: 2024-01-22 23:14
"""
import torch
import torch.nn as nn

class StridedConvolution(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=2, is_relu=True):
        super(StridedConvolution, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=1)
        self.relu = nn.ReLU(inplace=True)
        self.is_relu = is_relu

    def forward(self, x):
        x = self.conv(x)
        if self.is_relu:
            x = self.relu(x)
        return x

if __name__ == '__main__':
    input_data = torch.rand((1, 3, 64, 64))
    strided_conv = StridedConvolution(3, 64)
    output_data = strided_conv(input_data)
    print("Input shape:", input_data.shape)
    print("Output shape:", output_data.shape)



import numpy as np
from PIL import Image
import torch
import random
from thop import clever_format, profile
from torchsummary import summary

# def mixup(img1, labels, img2, labels2):
#     # Applies MixUp augmentation https://arxiv.org/pdf/1710.09412.pdf
#     r = np.random.beta(32.0, 32.0)  # mixup ratio, alpha=beta=32.0
#     img = (img1 * r + img2 * (1 - r)).astype(np.uint8)
#     labels = np.concatenate((labels, labels2), 0)
#     return img, labels
#
# def mix_up_single(batch_size, img, label, alpha=0.2):
#     def cir_shift(data):
#         index = list(range(1, batch_size)) + [0]
#         data = data[index, ...]
#         return data
#
#     lam = np.random.beta(alpha, alpha, batch_size)
#     lam_img = lam.reshape((batch_size, 1, 1, 1))
#     mix_img = lam_img * img + (1 - lam_img) * cir_shift(img)
#
#     lam_label = lam.reshape((batch_size, 1))
#     mix_label = lam_label * label + (1 - lam_label) * cir_shift(label)
#
#     return mix_img, mix_label
#
#
# def mix_up_muti(tmp, batch_size, img, label, alpha=0.2):
#     lam = np.random.beta(alpha, alpha, batch_size)
#     if tmp.is_first:
#         lam = np.ones(batch_size)
#         tmp.is_first = False
#
#     lam_img = lam.reshape((batch_size, 1, 1, 1))
#     mix_img = lam_img * img + (1 - lam_img) * tmp.image
#
#     lam_label = lam.reshape(batch_size, 1)
#     mix_label = lam_label * label + (1 - lam_label) * tmp.label
#     tmp.image = mix_img
#     tmp.label = mix_label
#
#     return mix_img, mix_label

def replicate(im, labels):
    # Replicate labels
    h, w = im.shape[:2]
    boxes = labels[:, 1:].astype(int)
    x1, y1, x2, y2 = boxes.T
    s = ((x2 - x1) + (y2 - y1)) / 2  # side length (pixels)
    for i in s.argsort()[:round(s.size * 0.5)]:  # smallest indices
        x1b, y1b, x2b, y2b = boxes[i]
        bh, bw = y2b - y1b, x2b - x1b
        yc, xc = int(random.uniform(0, h - bh)), int(random.uniform(0, w - bw))  # offset x, y
        x1a, y1a, x2a, y2a = [xc, yc, xc + bw, yc + bh]
        im[y1a:y2a, x1a:x2a] = im[y1b:y2b, x1b:x2b]  # im4[ymin:ymax, xmin:xmax]
        labels = np.append(labels, [[labels[i, 0], x1a, y1a, x2a, y2a]], axis=0)

    return im, labels

def preprocess_input(image):
    """
    将图像归一化到 [0, 1] 的范围
    :param image: 输入的图像
    :param open: 是否需要打开图像文件（默认为 False）
    :return: 归一化后的图像
    """
    if image.max() > 1:
        normalized_image = np.asarray(image) / 255.0
        return normalized_image
    else:
        return image


def resizepad_image(image, size, frame=True):
    """
    将调整图像大小并进行灰度填充
    :param image: 输入图像, PIL Image 对象
    :param size: 目标尺寸，形如 (width, height)
    :param frame: 是否进行不失真的resize
    :return: 调整大小后的图像，PIL Image 对象
    """
    iw, ih = image.size
    w, h = size
    if frame:
        scale = min(w/iw, h/ih)
        nw = int(iw*scale)
        nh = int(ih*scale)

        image = image.resize((nw,nh), Image.BICUBIC)
        new_image = Image.new('RGB', size, (128,128,128))
        new_image.paste(image, ((w-nw)//2, (h-nh)//2))
        return new_image, nw, nh
    else:
        new_image = image.resize((w, h), Image.BICUBIC)
        return new_image

def SeedEvery(seed=11, rank=0):
    """
    :param seed: 设置基础随机种子
    :param rank: 进程的排名，用于为每个进程设置不同的种子
    """
    combined_seed = seed + rank
    random.seed(combined_seed)
    np.random.seed(combined_seed)
    torch.manual_seed(combined_seed)
    if torch.cuda.is_available():
        import torch.backends.cudnn as cudnn
        import torch.backends.cuda as cuda
        cuda.matmul.allow_tf32 = True
        cudnn.benchmark = True
        torch.cuda.manual_seed(combined_seed)
        torch.cuda.manual_seed_all(combined_seed)

def summarys(input_shape, model):
    """
    打印模型的摘要信息，并计算模型的总浮点运算量和总参数数量
    :param input_shape:
    :param model:要进行计算的模型
    """
    device = "cuda" if torch.cuda.is_available() else "cpu"
    models = model.to(device)
    summary(models, (3, input_shape[0], input_shape[1]))

    dummy_input = torch.randn(1, 3, input_shape[0], input_shape[1]).to(device)
    flops, params = profile(models.to(device), (dummy_input, ), verbose=False)
    flops = flops * 2
    flops, params = clever_format([flops, params], "%.3f")
    print('Total GFLOPS: %s' % (flops))
    print('Total params: %s' % (params))





"""
Copyright (c) 2023, Auorui.
All rights reserved.

大多是考研数学所用的,闲来无事写的
"""
from .point_line import *
from .compute import *
from .arithmetic import *
from .constant import *
from .function import *
from .Numpy import *



"""
- author: Auorui (夏天是冰红茶)
- creation time:2022.10
- pyzjr is the Codebase accumulated by my python programming.
  At present, it is only for my personal use. If you want to
  use it, please contact me. Here are my email and WeChat.
- WeChat: z15583909992
- Email: zjricetea@gmail.com / 2165648225@qq.com
- Note: Currently still being updated, please refer to the latest version for any changes that may occur

    https://www.pepy.tech/projects/pyzjr
    API documentation website:
"""
from pyzjr.utils import *

from pyzjr.Z import * 
from pyzjr.Math import *
from pyzjr.PIC import *
from pyzjr.data import *
from pyzjr.devices import *

from pyzjr.augmentation import *
from pyzjr.mediapipe import *
from pyzjr.dlearn import *
from pyzjr.core import *
from pyzjr.videos import *
from pyzjr.version import __version__

from pyzjr.nn import *
from pyzjr.Models import *


# 学习率动态调整
import torch
import torch.optim as optim
import math
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from pyzjr.core import _LRScheduler

__all__ = [
            # 项目常用的两种
           "get_optimizer",
           "CustomScheduler",
           # 复现 https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate torch官方实现
           "FixedStepLR",
           "MultiStepLR",
           "CosineAnnealingLR",
           "WarmUpLR",
            # 继承了torch.optim.lr_scheduler的_LRScheduler
           "WarmUp",
           "FindLR",
            # 小工具：用于学习率范围测试
           "lr_finder",]

def get_optimizer(model, optimizer_type='adam', init_lr=0.001, momentum=None, weight_decay=None):
    """
    根据指定的优化器类型返回相应的优化器对象，并根据批次大小调整初始学习率。

    :param model: 要优化的神经网络模型
    :param optimizer_type: 优化器类型，可以是 'adam' 或 'sgd'，默认为 'adam'
    :param init_lr: 初始学习率，默认为 0.001
    :param momentum: SGD优化器的动量参数，默认为 None
    :param weight_decay: 权重衰减（L2正则化）参数，默认为 None
    :return: 优化器对象
    """
    if optimizer_type == 'adam':
        optimizer = optim.Adam(model.parameters(), lr=init_lr, betas=(momentum or 0.9, 0.999),
                               weight_decay=(weight_decay or 1e-4))
    elif optimizer_type == 'sgd':
        optimizer = optim.SGD(model.parameters(), lr=init_lr, momentum=(momentum or 0.9), nesterov=True,
                              weight_decay=(weight_decay or 1e-4))
    else:
        raise ValueError("Unsupported optimizer type:: {}".format(optimizer_type))

    return optimizer

class CustomScheduler:
    def __init__(self, optimizer, num_step, epochs, mode='cos', warmup=True, warmup_epochs=1, warmup_factor=1e-3, milestones=(30, 60, 90), gamma=0.1):
        """
        自定义学习率调度器，支持余弦退火（'cos'）和多步衰减（'step'）两种学习率调整模式。

        :param optimizer: 优化器对象，用于更新模型参数。
        :param num_step: 训练步数总数，通常等于 len(train_loader) * epochs。
        :param epochs: 总训练周期数。
        :param mode: 学习率调整模式，可以是 'cos'（余弦退火）或 'step'（多步衰减）。默认为 'cos'。
        :param warmup: 是否使用学习率预热。默认为 True。
        :param warmup_epochs: 学习率预热的周期数。仅在 warmup 为 True 时有效。默认为 1。
        :param warmup_factor: 学习率预热的初始倍率因子。仅在 warmup 为 True 时有效。默认为 1e-3。
        :param milestones: 多步衰减模式下的学习率降低的里程碑 epoch 列表。仅在 mode 为 'step' 时有效。默认为 (30, 60, 90)。
        :param gamma: 多步衰减模式下的学习率降低倍率因子。仅在 mode 为 'step' 时有效。默认为 0.1。
        """
        self.optimizer = optimizer
        self.num_step = num_step
        self.epochs = epochs
        self.use_multi_step = False
        if warmup is False:
            warmup_epochs = 0
        if mode == "cos":
            self.warmup = warmup
            self.warmup_epochs = warmup_epochs
            self.warmup_factor = warmup_factor
        elif mode == "step":
            self.use_multi_step = True
            self.milestones = milestones
            self.gamma = gamma
        self.scheduler = self.get_scheduler()

    def get_scheduler(self):
        assert self.num_step > 0 and self.epochs > 0
        if self.use_multi_step:
            return optim.lr_scheduler.MultiStepLR(self.optimizer, milestones=self.milestones, gamma=self.gamma)
        else:
            def f(x):
                """
                根据step数返回一个学习率倍率因子，
                """
                if self.warmup is True and x <= (self.warmup_epochs * self.num_step):
                    alpha = float(x) / (self.warmup_epochs * self.num_step)

                    return self.warmup_factor * (1 - alpha) + alpha
                else:
                    # warmup后lr倍率因子从1 -> 0
                    # 参考deeplab_v2: Learning rate policy
                    return (1 - (x - self.warmup_epochs * self.num_step) / ((self.epochs - self.warmup_epochs) * self.num_step)) ** 0.9

            return optim.lr_scheduler.LambdaLR(self.optimizer, lr_lambda=f)

    def step(self):
        self.scheduler.step()

    def get_lr(self):
        return [group['lr'] for group in self.optimizer.param_groups]


class FindLR(_LRScheduler):
    """
    exponentially increasing learning rate

    Args:
        optimizer: optimzier(e.g. SGD)
        num_iter: totoal_iters
        max_lr: maximum  learning rate
    """
    def __init__(self, optimizer, max_lr=10, num_iter=100, last_epoch=-1):

        self.total_iters = num_iter
        self.max_lr = max_lr
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        return [base_lr * (self.max_lr / base_lr) ** (self.last_epoch / (self.total_iters + 1e-32)) for base_lr in self.base_lrs]


class lr_finder():
    def __init__(self,net, dataloader, loss_function, optimizer_type="sgd",num_iter=100,batch_size=4):
        self.net = net
        self.dataloader = dataloader
        self.loss_function = loss_function
        self.optimizer_type = optimizer_type
        self.num_iter = num_iter
        self.batch_size = batch_size

    def update(self, init_lr=1e-7, max_lr=10):
        n = 0
        learning_rate = []
        losses = []
        optimizer = get_optimizer(self.net, self.optimizer_type, init_lr)
        lr_scheduler = FindLR(optimizer, max_lr=max_lr, num_iter=self.num_iter)
        epoches = int(self.num_iter / len(self.dataloader)) + 1

        for epoch in range(epoches):
            self.net.train()
            for batch_index, (images, labels) in enumerate(self.dataloader):
                if n > self.num_iter:
                    break
                if torch.cuda.is_available():
                    images = images.cuda()
                    labels = labels.cuda()

                optimizer.zero_grad()
                predicts = self.net(images)
                loss = self.loss_function(predicts, labels)
                if torch.isnan(loss).any():
                    n += 1e8
                    break
                loss.backward()
                optimizer.step()
                lr_scheduler.step()

                print('Iterations: {iter_num} [{trained_samples}/{total_samples}]\tLoss: {:0.4f}\tLR: {:0.8f}'.format(
                    loss.item(),
                    optimizer.param_groups[0]['lr'],
                    iter_num=n,
                    trained_samples=batch_index * self.batch_size + len(images),
                    total_samples=len(self.dataloader),
                ))

                learning_rate.append(optimizer.param_groups[0]['lr'])
                losses.append(loss.item())
                n += 1

        self.learning_rate = learning_rate[10:-5]
        self.losses = losses[10:-5]

    def plotshow(self, show=True):
        import matplotlib
        matplotlib.use("TkAgg")
        fig, ax = plt.subplots(1, 1)
        ax.plot(self.learning_rate, self.losses)
        ax.set_xlabel('learning rate')
        ax.set_ylabel('losses')
        ax.set_xscale('log')
        ax.xaxis.set_major_formatter(plt.FormatStrFormatter('%.0e'))
        if show:
            plt.show()

    def save(self, path='result.jpg'):
        self.plotshow(show=False)
        plt.savefig(path)

class FixedStepLR:
    """
    Compared with Torch's official StepLR and completed the functional implementation.
    每 step_size 个周期通过 gamma 衰减每个参数组的学习率。
    """
    def __init__(self, optimizer, step_size, gamma=0.1):
        self.optimizer = optimizer
        self.step_size = step_size
        self.gamma = gamma
        self.last_epoch = 0

    def get_lr(self):
        if self.last_epoch == 0 or (self.last_epoch % self.step_size != 0):
            return [group['lr'] for group in self.optimizer.param_groups]
        else:
            return [lr * self.gamma for lr in [group['lr'] for group in self.optimizer.param_groups]]

    def step(self):
        self.last_epoch += 1
        new_lr = self.get_lr()
        for param_group, lr in zip(self.optimizer.param_groups, new_lr):
            param_group['lr'] = lr


class MultiStepLR:
    """
    Compared with Torch's official MultiStepLR and completed the functional implementation.
    当达到里程碑之一, 将每个参数组的学习率衰减 gamma.
    """
    def __init__(self, optimizer, milestones, gamma=0.1):
        self.optimizer = optimizer
        self.milestones = self.to_constant(milestones)
        self.gamma = gamma
        self.last_epoch = 0

    def to_constant(self, x):
        if isinstance(x, list):
            return [int(element) for element in x]

    def get_lr(self):
        import bisect
        milestone_index = bisect.bisect_right(self.milestones, self.last_epoch)
        if self.last_epoch not in self.milestones:
            return [group['lr'] for group in self.optimizer.param_groups]
        else:
            return [group['lr'] * self.gamma ** milestone_index
                    for group in self.optimizer.param_groups]

    def step(self):
        self.last_epoch += 1
        new_lr = self.get_lr()
        for param_group, lr in zip(self.optimizer.param_groups, new_lr):
            param_group['lr'] = lr

class CosineAnnealingLR:
    def __init__(self, optimizer, T_max, eta_min=0):
        self.optimizer = optimizer
        self.T_max = T_max
        self.eta_min = eta_min
        self.last_epoch = 0
        if self.last_epoch == 0:
            for group in optimizer.param_groups:
                group.setdefault('initial_lr', group['lr'])
        self.base_lrs = list(map(lambda group: group['initial_lr'], optimizer.param_groups))

    def get_lr(self):
        if self.last_epoch == 0:
            return self.base_lrs
        elif (self.last_epoch - 1 - self.T_max) % (2 * self.T_max) == 0:
            return [group['lr'] + (base_lr - self.eta_min) *
                    (1 - math.cos(math.pi / self.T_max)) / 2
                    for base_lr, group in zip(self.base_lrs, self.optimizer.param_groups)]

        return [(1 + math.cos(math.pi * self.last_epoch / self.T_max)) /
                (1 + math.cos(math.pi * (self.last_epoch - 1) / self.T_max)) *
                (group['lr'] - self.eta_min) + self.eta_min
                for group in self.optimizer.param_groups]

    def step(self):
        self.last_epoch += 1
        new_lr = self.get_lr()
        for param_group, lr in zip(self.optimizer.param_groups, new_lr):
            param_group['lr'] = lr

class WarmUp(_LRScheduler):
    """
    warmup_training learning rate scheduler
    Args:
        optimizer: optimzier
        total_iters: totoal_iters of warmup phase
    """
    def __init__(self, optimizer, total_iters, last_epoch=-1):
        self.total_iters = total_iters
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        """we will use the first m batches, and set the learning
        rate to base_lr * m / total_iters
        """
        return [base_lr * self.last_epoch / (self.total_iters + 1e-8) for base_lr in self.base_lrs]

class WarmUpLR:
    def __init__(self, optimizer, train_loader, epochs, warmup_epochs=1, warmup_factor=1e-3,):
        """
        :param optimizer: 优化器对象，用于更新模型参数。
        :param train_loader: 训练的数据加载器
        :param epochs: 总训练周期数。
        :param warmup_epochs: 学习率预热的周期数。仅在 warmup 为 True 时有效。默认为 1。
        :param warmup_factor: 学习率预热的初始倍率因子。仅在 warmup 为 True 时有效。默认为 1e-3。
        """
        self.optimizer = optimizer
        self.warmup_epochs = warmup_epochs
        self.epochs = epochs
        self.num_step = len(train_loader) * self.epochs   # num_step训练步数总数，通常等于 len(train_loader) * epochs。
        self.warmup_factor = warmup_factor
        self.scheduler = self.get_scheduler()

    def get_scheduler(self):
        def f(x):
            """
            根据step数返回一个学习率倍率因子，
            """
            if self.warmup_epochs > 0 and x <= (self.warmup_epochs * self.num_step):
                alpha = float(x) / (self.warmup_epochs * self.num_step)
                return self.warmup_factor * (1 - alpha) + alpha
            else:
                # warmup后lr倍率因子从1 -> 0
                # 参考deeplab_v2: Learning rate policy
                return (1 - (x - self.warmup_epochs * self.num_step) / ((self.epochs - self.warmup_epochs) * self.num_step)) ** 0.9
        return optim.lr_scheduler.LambdaLR(self.optimizer, lr_lambda=f)

    def step(self):
        self.scheduler.step()

    def get_lr(self):
        return [group['lr'] for group in self.optimizer.param_groups]

if __name__=="__main__":
    from torch.optim import lr_scheduler

    class SimpleModel(torch.nn.Module):
        def __init__(self):
            super(SimpleModel, self).__init__()
            self.linear = torch.nn.Linear(1, 1)

        def forward(self, x):
            return self.linear(x)

    x_data = torch.tensor([[1.0], [2.0], [3.0]])
    y_data = torch.tensor([[2.0], [4.0], [6.0]])

    model = SimpleModel()
    optimizer = optim.SGD(model.parameters(), lr=0.1)

    criterion = torch.nn.MSELoss()
    # 固定的步数gamma衰减测试
    # scheduler = lr_scheduler.StepLR(optimizer, step_size=30, gamma=0.1)
    # scheduler = FixedStepLR(optimizer, step_size=30, gamma=0.1)
    # 当达到里程碑时进行gamma衰减测试
    # scheduler = lr_scheduler.MultiStepLR(optimizer, milestones=[30, 80], gamma= 0.1)
    # scheduler = MultiStepLR(optimizer, milestones=[30, 80], gamma=0.1)
    # scheduler = lr_scheduler.CosineAnnealingLR(optimizer, T_max=25)
    scheduler = WarmUpLR(optimizer, num_step=100, epochs=100)
    # scheduler = lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=25, T_mult=1, eta_min=0)
    epochs = 100
    for epoch in range(epochs):
        y_pred = model(x_data)
        loss = criterion(y_pred, y_data)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        scheduler.step()
        print(f'Epoch {epoch + 1}/{epochs}, Loss: {loss.item()}, Learning Rate: {optimizer.param_groups[0]["lr"]}')

    with torch.no_grad():
        test_data = torch.tensor([[4.0]])
        predicted_y = model(test_data)
        print(f'After training, input 4.0, predicted output: {predicted_y.item()}')
